#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **non-pooled worker thread** (i.e., platform-portable, pure-Qt,
:class:`QThread`-based thread wrapper running zero or more subordinate
:class:`QBetseeLoneThreadWorkerABC`-based workers at a given time) classes.
'''

#FIXME: Most (all?) methods defined below are intended to be run *ONLY* from the
#main thread and hence should raise "BetseePySideThreadException" when this is
#*NOT* the case. At the very least, this should be well-documented below.

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, QThread
# from betse.util.io.log import logs
from betse.util.type.types import type_check  #, CallableTypes
from betsee.guiexception import BetseePySideThreadException
from betsee.util.thread.lone.guiloneworker import QBetseeLoneThreadWorkerABC

# ....................{ SUPERCLASSES                       }....................
class QBetseeLoneThread(QThread):
    '''
    Low-level **non-pooled worker thread"** (i.e., platform-portable, pure-Qt,
    :class:`QThread`-based wrapper running zero or more subordinate
    :class:`QBetseeLoneThreadWorkerABC`-based workers at a given time).

    Caveats
    ----------
    This obsolete superclass has been superceded by the superior
    :class:`betse.util.thread.pool.guipoolthread.QBetseeThreadPool` superclass,
    whose :class:`QRunnable`-based API requires substantially less boilerplate.

    The *only* methods of this subclass safely callable by external objects are
    the superclass :meth:`start` method and all subclass-specific public methods
    and properties. Other superclass methods (e.g., :meth:`exec`, :meth:`exit`,
    :meth:`quit`, :meth:`run`) should be considered unsafe for external usage
    and hence *never* called externally.

    Design
    ----------
    By design, this subclass intentionally does *not* override the superclass
    :meth:`run` method. While technically feasible (and even sadly encouraged by
    official documentation), doing so invites practical complications that
    *cannot* be trivially circumvented. These include:

    * The inability to run signalled slots of this subclass in this thread.
      Instead, each such slot is run in the thread from which that signal was
      emitted (e.g., the main event-handling thread). This severe constraint
      prohibits this thread from handling signals received from other threads,
      effectively rendering this thread useless for *most* real-world purposes.
    * The need to manually "spin up" a Qt event loop by calling the superclass
      :meth:`exec` method within the subclass :meth:`run` implementation.
    * The need to manually handle the threading lifecycle (e.g., pausing,
      resuming) with error-prone, low-level Qt primitives, including:
      * Mutual exclusions (e.g., :class:`QMutex`, :class:`QSemaphore`).
      * Thread-safe types (e.g., C++-style atomics).

    Instead, this subclass *strongly* recommends that
    :class:`QBetseeLoneThreadWorkerABC`-based workers be moved from the main
    event-handling thread to this thread by the :meth:`adopt_worker` method.

    See Also
    ----------
    https://mayaposch.wordpress.com/2011/11/01/how-to-really-truly-use-qthreads-the-full-explanation
        Blog article on the inadvisability of overriding the :meth:`run` method.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this worker thread.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Garbage collect all child objects of this parent thread *AFTER* this
        # thread's event loop and hence execution gracefully terminates (e.g.,
        # by an external call to the halt() method).
        self.finished.connect(self.deleteLater)

    # ..................{ PROPERTIES                         }..................
    @property
    def process_name(self) -> str:
        '''
        Name of the OS-level process associated with this thread, equivalent to
        the Qt-specific name of this object.

        While Qt-specific object names are typically hidden from end users, the
        name of this thread is perhaps surprisingly published to end users
        through the process list for most operating systems (e.g., via the
        external ``ps -L`` command on POSIX-compatible systems).

        This name defaults to the unqualified and largely unreadable name of
        this class (e.g., ``QBetseeLoneThread``). Callers are advised to set
        this property to a human-readable name uniquely identifying this thread
        *before* calling the :meth:`start` method. (Attempting to set this name
        *after* calling that method silently reduces to a noop.)
        '''

        return self.objectName()


    @process_name.setter
    @type_check
    def process_name(self, process_name: str) -> None:
        '''
        Set the name of the OS-level process associated with this thread,
        equivalent to the Qt-specific name of this object, to the passed string.
        '''

        self.setObjectName(process_name)

    # ..................{ HALTERS                            }..................
    @type_check
    def halt(self) -> None:
        '''
        Gracefully terminate this thread.

        Specifically, this method:

        * Gracefully terminates this thread's event loop (i.e., the
          :meth:`exec` method) and hence this thread's execution (i.e., the
          :meth:`run` method).
        * Waits indefinitely for this thread's execution to gracefully
          terminate.

        Caveats
        ----------
        This method does *not* gracefully terminate any adopted worker of this
        thread running in this thread at the time of this call. Rather, such a
        worker will be non-gracefully terminated.

        The :meth:`terminate` method non-gracefully terminating this thread
        should *never* be called for any reason whatsoever. Doing so typically
        disrupts sane thread-safety in a manner inducing application-terminating
        segmentation faults. That's bad.

        The :meth:`exit` method gracefully terminating this thread's event loop
        should *never* be called either. This method's name is poorly chosen and
        bears no relation to actually exiting, quitting, or otherwise
        terminating this thread. Instead, this method merely undoes any prior
        call to the :meth:`exec` method starting this thread's event loop.

        The :meth:`quit` method is a perfect alias of the :meth:`exit` method
        and hence equally inadvisable for external use. Don't call it, either.

        In short, the :class:`QThread` API is a contemptible mess. Avoid it
        wherever possible; complain about it on StackOverflow and Reddit
        wherever impossible.
        '''

        # Gracefully terminate this thread's event loop and hence this thread's
        # execution. This should be the *ONLY* call to this method across the
        # entire codebase.
        self.quit()

        # Wait indefinitely for this thread to gracefully terminate.
        self.wait()

    # ..................{ ADOPTERS                           }..................
    @type_check
    def adopt_worker(self, *workers: QBetseeLoneThreadWorkerABC) -> None:
        '''
        Adopt all passed application-specific workers into this thread.

        Specifically, this method iteratively:

        * Moves each such worker into this thread.
        * Automates the lifecycle of these workers by deleting these workers
          when this thread is deleted.

        This method does *not* classify these workers as instance variables of
        this object. This thread object is technically *not* a thread; this
        thread object is merely a high-level wrapper encapsulating a thread.
        This thread object does *not* reside in this thread but rather in the
        parent thread (e.g., main event-handling thread) in which this object
        was instantiated. This worker *does* reside in this thread after this
        method returns, however. Objects residing in different threads should
        *never* retain references to one another. Ergo, the conclusion follows.

        Caveats
        ----------
        Avoid manually calling the :meth:`QBetseeLoneThreadWorkerABC.moveToThread`
        method to move workers into this thread. Doing so prevents this method
        from automating the lifecycle of these workers -- notably, the guarantee
        that these workers be deleted when this thread is deleted.

        Parameters
        ----------
        workers : tuple[QBetseeLoneThreadWorkerABC]
            Tuple of all workers to be adopted by this thread.
        '''

        # For each such worker...
        for worker in workers:
            # Move this worker into this thread.
            worker.moveToThread(self)

            #FIXME: *NOPE.* While this seems sensible, actually doing this
            #results in the entire application fatally segmentation-faulting
            #without a human-readable exception or error. Ergo, bad. Somehow.

            # Garbage collect all child objects of this worker *AFTER* this
            # thread's event loop and hence execution gracefully terminates
            # (e.g., by an external call to the halt() method).
            # self.finished.connect(worker.deleteLater)

# ....................{ SUPERCLASSES ~ single              }....................
class QBetseeLoneSingleThread(QBetseeLoneThread):
    '''
    Low-level **non-pooled single worker thread"** (i.e., platform-portable,
    pure-Qt, :class:`QThread`-based thread wrapper, confined to run a single
    subordinate :class:`QBetseeLoneThreadWorkerABC`-based worker at a given
    time).

    By default, both the standard :class:`QThread` class and our
    :class:`QBetseeLoneThread` superclass implicitly allow a one-to-many
    relation between a parent thread and its child workers such that a single
    thread may run an arbitrary number of (i.e., zero or more) workers.

    Attributes
    ----------
    _is_worker : bool
        ``True`` only if this thread has adopted a worker (i.e., the
        :meth:`adopt_worker` method has been called more recently than the
        :meth:`unadopt_worker` method).
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this single worker thread.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # By default, this thread has no adopted worker.
        self._is_worker = False

    # ..................{ PROPERTIES                         }..................
    # Read-only properties.

    @property
    def is_worker(self) -> bool:
        '''
        ``True`` only if this thread has already adopted a worker (i.e., the
        :meth:`adopt_worker` method has been called more recently than the
        :meth:`unadopt_worker` method).
        '''

        return self._is_worker

    # ..................{ EXCEPTIONS                         }..................
    def _die_if_worker(self) -> None:
        '''
        Raise an exception if this thread has already adopted a worker.

        Raises
        ----------
        BetseePySideThreadException
            If this thread has already adopted a worker.

        See Also
        ----------
        :meth:`is_worker`
            Further details.
        '''

        # If this thread has already adopted a worker, raise an exception.
        if self.is_worker:
            raise BetseePySideThreadException(QCoreApplication.translate(
                'QBetseeWorkerSingleThread',
                'Worker already adopted by this thread.'))


    def _die_unless_worker(self) -> None:
        '''
        Raise an exception if unless thread has already adopted a worker (i.e.,
        if this thread has adopted no worker).

        Raises
        ----------
        BetseePySideThreadException
            If this thread has adopted no worker.

        See Also
        ----------
        :meth:`is_worker`
            Further details.
        '''

        # If this thread has adopted no worker, raise an exception.
        if not self.is_worker:
            raise BetseePySideThreadException(QCoreApplication.translate(
                'QBetseeWorkerSingleThread',
                'No worker adopted by this thread.'))

    # ..................{ HALTERS                            }..................
    @type_check
    def halt(self) -> None:
        '''
        Gracefully terminate this thread.

        Specifically, this method:

        * If this thread has adopted a worker, unadopts this worker. Doing so
          ensures exceptions are raised in the common edge case that this worker
          is still running.
        * Gracefully terminates this thread's event loop (i.e., the
          :meth:`exec` method) and hence this thread's execution (i.e., the
          :meth:`run` method).
        * Waits indefinitely for this thread's execution to gracefully
          terminate.

        Caveats
        ----------
        This method does *not* gracefully terminate any adopted worker of this
        thread running in this thread at the time of this call. Rather, such a
        worker will be non-gracefully terminated.

        The :meth:`terminate` method non-gracefully terminating this thread
        should *never* be called for any reason whatsoever. Doing so typically
        disrupts sane thread-safety in a manner inducing application-terminating
        segmentation faults. That's bad.

        The :meth:`exit` method gracefully terminating this thread's event loop
        should *never* be called either. This method's name is poorly chosen and
        bears no relation to actually exiting, quitting, or otherwise
        terminating this thread. Instead, this method merely undoes any prior
        call to the :meth:`exec` method starting this thread's event loop.

        The :meth:`quit` method is a perfect alias of the :meth:`exit` method
        and hence equally inadvisable for external use. Don't call it, either.

        In short, the :class:`QThread` API is a contemptible mess. Avoid it
        wherever possible; complain about it on StackOverflow and Reddit
        wherever impossible.
        '''

        # If this thread has adopted a worker, unadopt this worker *BEFORE*
        # calling the superclass method, which terminates this thread. Doing so
        # ensures exceptions are raised in the common edge case that this worker
        # is still running.
        if self.is_worker:
            self.unadopt_worker()

        # Gracefully terminate this thread.
        super().halt()

    # ..................{ ADOPTERS                           }..................
    @type_check
    def adopt_worker(self, *workers: QBetseeLoneThreadWorkerABC) -> None:
        '''
        Adopt all passed application-specific workers into this thread if no
        other worker has already been adopted by this thread *or* raise an
        exception otherwise (i.e., if this thread has already adopted a worker).

        Specifically, this method iteratively:

        * Moves each such worker into this thread while preserving the
          guarantees contractually stipulated by this thread -- notably, that
          this thread run one and *only* one worker at a time.

        Design
        ----------
        For safety, this method does *not* classify this worker as an instance
        variable of this object. Why? Because this object is technically *not* a
        thread; this object is merely a high-level wrapper encapsulating a
        thread. This object does *not* reside in this thread but rather in the
        parent thread (e.g., main event-handling thread) in which this object
        was instantiated. This worker *does* reside in this thread after this
        method returns, however. Objects residing in different threads should
        *never* retain references to one another. Ergo, the conclusion follows.

        Instead, this method merely classifies an arbitrary worker to have now
        been adopted by this thread via the :meth:`is_worker` property.

        Caveats
        ----------
        The :meth:`QBetseeLoneThreadWorkerABC.moveToThread` method should
        *never* be called to manually move a worker into this thread. Doing so
        violates subclass encapsulation and hence the guarantee that this thread
        run *only* a single worker at a time.
        '''

        # If this thread has already adopted a worker, raise an exception.
        self._die_if_worker()

        # Move this worker to this thread.
        super().adopt_worker(workers)

        #FIXME: Connect appropriate signals and slots... after defining them. In
        #particular, at least:
        #
        #* Augment the unadopt_worker() method into a slot.
        #* Connect the worker.halt() signal to the unadopt_worker() slot.

        # Record this thread as having a worker *AFTER* successfully doing so.
        self._is_worker = True


    #FIXME: This method should explicitly handle the common edge case of an
    #adopted worker that is still running by detecting this edge case and
    #either:
    #
    #* Raising an exception. (Non-ideal, but acceptable.)
    #* Emitting a signal connected to the halt() slot of this worker,
    #  instructing this worker to gracefully shutdown. While preferable,
    #  implementing this logic is presumably complicated by the need for this
    #  thread to wait for this slot to return before terminating this thread.
    def unadopt_worker(self) -> None:
        '''
        Unadopt the passed application-specific worker from this thread if a
        worker has already been adopted by this thread *or* raise an exception
        otherwise (i.e., if this thread has adopted no such worker).
        '''

        # If this thread has adopted no such worker, raise an exception.
        self._die_unless_worker()

        #FIXME: Do something appropriate here. Maybe emit appropriate signals?

        # Record this thread as having no worker *AFTER* successfully doing so.
        self._is_worker = False



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **simulator phase** (i.e., simulation phase to be queued for
modelling and/or exporting by this simulator) functionality.
'''

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QObject, Signal, Slot  # QCoreApplication,
from betse.science.phase.phaseenum import SimPhaseKind
from betse.util.io.log import logs
from betse.util.type import enums
from betse.util.type.decorator.decmemo import property_cached
from betse.util.type.types import type_check  #, CallableTypes
from betsee.gui.simtab.run.guisimrunabc import QBetseeSimmerStatefulABC
from betsee.gui.simtab.run.guisimrunenum import SimmerState
from betsee.gui.simtab.run.guisimrunstate import (
    SIMMER_STATE_TO_PHASE_STATUS,
    SIMMER_STATES_FROM_FLUID,
)
from betsee.gui.window.guimainwindow import QBetseeMainWindow

# ....................{ SUPERCLASSES                      }....................
class QBetseeSimmerPhase(QBetseeSimmerStatefulABC):
    '''
    **Simulator phase controller** (i.e., :mod:`PySide2`-based object wrapping
    a simulator phase to be queued for modelling and/or exporting by this
    simulator).

    This controller maintains all state required to interactively manage this
    simulator phase.

    Attributes (Private: Non-widgets)
    ----------
    _kind : SimPhaseKind
        Type of simulation phase controlled by this controller.

    Attributes (Private: Widgets)
    ----------
    _queue_modelling_lock : QCheckBox
        Checkbox toggling whether this phase is queueable for modelling.
    _queue_modelling : QCheckBox
        Checkbox toggling whether this phase is queued for modelling.
    _queue_exporting : QCheckBoxOrNoneTypes
        Checkbox toggling whether this phase is queued for exporting if this
        phase supports exporting *or* ``None`` otherwise. While most phases
        support exporting, some (e.g., the seed phase) do *not*.
    _status : QLabel
        Label synopsizing the current state of this phase.
    '''

    # ..................{ SIGNALS                           }..................
    queue_changed = Signal(QObject)
    '''
    Signal emitted by the :meth:`_toggle_queue_subkind` slot immediately after
    changing the state of this simulator phase to be either
    :attr:`SimmerState.QUEUED` or :attr:`SimmerState.UNQUEUED`, passed this
    simulator phase itself.

    This signal enables the parent simulator proactor to respond to user
    requests to enqueue or dequeue any child simulator phase, which are
    sufficiently unique events to warrant special-case handling.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this simulator phase.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all remaining instance variables for safety.
        self._kind = None
        self._queue_modelling_lock = None
        self._queue_modelling = None
        self._queue_exporting = None
        self._status = None


    @type_check
    def init(
        self,
        main_window: QBetseeMainWindow,
        kind: SimPhaseKind,
    ) -> None:
        '''
        Finalize the initialization of this simulator phase, owned by the
        passed main window widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of this simulator.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        kind : SimPhaseKind:
            Type of simulation phase controlled by this controller.
        '''

        # Initialize our superclass with the passed window.
        super().init(main_window)

        # Classify all remaining passed parameters.
        self._kind = kind

        # Log this initialization *AFTER* classifying these parameters, as
        # required by the "name" property embedded in this log message.
        logs.log_debug(
            'Sanitizing simulator phase "%s" state...', self.name)

        # Initialize all widgets concerning simulator state.
        self._init_widgets(main_window)

        # Connect all relevant signals and slots *AFTER* initializing these
        # widgets, as the former typically requires the latter.
        self._init_connections(main_window)


    @type_check
    def _init_widgets(self, main_window: QBetseeMainWindow) -> None:
        '''
        Create all widgets owned directly by this object *and* initialize all
        other widgets (*not* always owned by this object) concerning this
        simulator phase.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow`
            widget.
        '''

        # Names of all instance variables of this main window yielding widgets
        # specific to this simulator phase.
        is_unqueueable_model_name = 'sim_run_queue_{}_model_lock'.format(
            self.name)
        is_queued_model_name  = 'sim_run_queue_{}_model' .format(self.name)
        is_queued_export_name = 'sim_run_queue_{}_export'.format(self.name)
        status_name           = 'sim_run_queue_{}_status'.format(self.name)

        # Classify mandatory widgets unconditionally defined for *ALL* phases.
        self._queue_modelling_lock = main_window.get_widget(
            widget_name=is_unqueueable_model_name)
        self._queue_modelling = main_window.get_widget(
            widget_name=is_queued_model_name)
        self._status = main_window.get_widget(
            widget_name=status_name)

        # Classify optional widgets conditionally defined for only some phases.
        self._queue_exporting = main_window.get_widget_or_none(
            widget_name=is_queued_export_name)


    @type_check
    def _init_connections(self, main_window: QBetseeMainWindow) -> None:
        '''
        Connect all relevant signals and slots of *all* widgets (including the
        main window, top-level widgets of that window, and leaf widgets
        distributed throughout this application) whose internal state pertains
        to the high-level state of this simulator phase.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow`
            widget.
        '''

        # Connect signals emitted by the checkbox queueing this phase for
        # modelling to corresponding slots, thus synchronizing the state of
        # this checkbox with both this phase and this phase's parent simulator.
        self._queue_modelling.toggled.connect(self._toggle_queue_subkind)

        # If this phase is exportable, connect signals emitted by the checkbox
        # queueing this phase for exporting to corresponding slots as above.
        if self._queue_exporting is not None:
            self._queue_exporting.toggled.connect(self._toggle_queue_subkind)

        # Connect signals emitted by widgets owned by this phase to slots
        # defined by this phase.
        self._queue_modelling_lock.toggled.connect(
            self._toggle_queue_modelling_lock)

        # Queue this phase to be modelled but *NOT* exported by default *AFTER*
        # connecting all signals emitted by this checkbox above.
        self._queue_modelling.setChecked(True)

    # ..................{ PROPERTIES ~ bool                 }..................
    # Read-only boolean properties.

    @property
    def is_queued(self) -> bool:
        '''
        ``True`` only if this simulator phase is currently queued (i.e., for
        modelling and/or exporting).
        '''

        return self.is_queued_modelling or self.is_queued_exporting


    # This trivial getter exists only for orthogonality with the corresponding
    # non-trivial is_queued_exporting() getter.
    @property
    def is_queued_modelling(self) -> bool:
        '''
        ``True`` only if this simulator phase is currently queued for
        modelling.
        '''

        return self._queue_modelling.isChecked()


    @property
    def is_queued_exporting(self) -> bool:
        '''
        ``True`` only if this simulator phase is currently queued for
        exporting.
        '''

        # Return true only if this phase is both...
        return (
            # Exportable.
            self._queue_exporting is not None and
            # Currently queued for exporting.
            self._queue_exporting.isChecked()
        )

    # ..................{ PROPERTIES ~ non-bool             }..................
    # Read-only non-boolean properties.

    @property
    def kind(self) -> SimPhaseKind:
        '''
        Type of simulation phase controlled by this controller.
        '''

        return self._kind


    @property_cached
    def name(self) -> str:
        '''
        Machine-readable alphabetic lowercase name of the type of simulation
        phase controlled by this controller (e.g., ``seed``, ``init``).
        '''

        return enums.get_member_name_lowercase(self._kind)

    # ..................{ SLOTS                             }..................
    @Slot(bool)
    def _toggle_queue_subkind(self, is_queued: bool) -> None:
        '''
        Slot signalled on either the user interactively *or* the codebase
        programmatically toggling the :class:`QCheckBox` widget corresponding
        to either the :attr:`_queue_modelling` *or* :attr:`_queue_exporting`
        variables.

        If this slot changes the state of this simulator phase to either
        :attr:`SimmerState.QUEUED` or :attr:`SimmerState.UNQUEUED`, this slot
        notifies interested parties of this fact by signalling the
        :attr:`queue_changed` with this simulator phase.

        Parameters
        ----------
        is_queued : bool
            ``True`` only if this :class:`QCheckBox` is currently checked.
        '''

        # Log this action.
        logs.log_debug(
            'Enqueueing simulator phase "%s"...' if is_queued else
            'Dequeueing simulator phase "%s"...', self.name)

        # If the current state of this phase is fluid (i.e., freely replaceable
        # with any other state)...
        if self.state in SIMMER_STATES_FROM_FLUID:
            # Reduce this state to whether this phase is queued or not.
            self.state = (
                SimmerState.QUEUED if self.is_queued else SimmerState.UNQUEUED)

            # Update the proactor's state to reflect this phase's state.
            self.queue_changed.emit(self)
        # Else, the current state of this phase is fixed and hence *NOT*
        # replaceable with any other state. For safety, this state is preserved
        # as is.
        #
        # Note that setting the "state" property above implicitly calls the
        # _update_state() method, which thus need (and indeed should) *NOT* be
        # explicitly recalled here. While technically safe, doing so would
        # nonetheless incur a senseless performance penalty.


    @Slot(bool)
    def _toggle_queue_modelling_lock(self, is_unqueueable_model: bool) -> None:
        '''
        Slot signalled on either the user interactively *or* the codebase
        programmatically toggling the checkable :class:`QToolButton` widget
        corresponding to the :attr:`_queue_modelling_lock` variable.

        Specifically, if:

        * This button is checked, this slot locks (i.e., disables) the
          :class:`QCheckBox` widget corresponding to the
          :attr:`_queue_modelling` variable.
        * This button is unchecked, this slot unlocks (i.e., enables) that
          :class:`QCheckBox` widget.

        Parameters
        ----------
        is_unqueueable_model : bool
            ``True`` only if this :class:`QToolButton` is currently checked.
        '''

        # One-liners for the greater glory of god-like efficiency.
        self._queue_modelling.setEnabled(not is_unqueueable_model)

    # ..................{ UPDATERS                          }..................
    def _update_state(self) -> None:

        # Text synopsizing the action being performed in this state *AFTER*
        # possibly setting this state above.
        status_text = SIMMER_STATE_TO_PHASE_STATUS[self.state]

        # Set the text of the label displaying this synopsis to this text.
        self._status.setText(status_text)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of all **clipboardable widget** (i.e., widget
transparently supporting copying, cutting, and pasting into and from the
platform-specific system clipboard) subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Qt, QEvent
from PySide2.QtGui import QKeyEvent
from PySide2.QtWidgets import QApplication
from betse.exceptions import BetseMethodUnimplementedException
from betse.util.type.obj import objects
# from betse.util.type.types import type_check

# ....................{ MIXINS                             }....................
# To avoid metaclass conflicts with the "QWidget" base class inherited by all
# widgets also inheriting this base class, this base class *CANNOT* be
# associated with another metaclass (e.g., "abc.ABCMeta").
class QBetseeClipboardWidgetMixin(object):
    '''
    Abstract base class of all **clipboardable widget** (i.e., widget
    transparently supporting copying, cutting, and pasting into and from the
    platform-specific system clipboard) subclasses.

    Most subclasses of this class support only plaintext and hence integrate
    only with the clipboard's plaintext buffer.

    Design
    ----------
    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be subclassed *first* rather than *last* in subclasses.
    '''

    # ..................{ SUBCLASS ~ mandatory : method      }..................
    # Subclasses are required to implement the following methods.

    def copy_selection_to_clipboard(self) -> None:
        '''
        Copy this widget's **current selection** (i.e., currently selected
        subset of this widget's value(s)) to the system clipboard, silently
        replacing the prior contents if any.
        '''

        raise BetseMethodUnimplementedException()


    def cut_selection_to_clipboard(self) -> None:
        '''
        **Cut** (i.e., copy and then remove as a single atomic operation) the
        this widget's **current selection** (i.e., currently selected subset of
        this widget's value(s)) to the system clipboard, silently replacing the
        prior contents if any.
        '''

        raise BetseMethodUnimplementedException()


    def paste_clipboard_to_selection(self) -> None:
        '''
        Paste the contents of the system clipboard over this widget's **current
        selection** (i.e., currently selected subset of this widget's value(s)),
        silently replacing the prior selection if any.
        '''

        raise BetseMethodUnimplementedException()


class QBetseeClipboardScalarWidgetMixin(QBetseeClipboardWidgetMixin):
    '''
    Abstract base class of all **scalar clipboardable widget** (i.e., scalar
    :mod:`PySide2` widget transparently supporting copying, cutting, and pasting
    into and from the platform-specific system clipboard) subclasses.

    In this context, the term "scalar" encompasses all widget subclasses whose
    contents reduce to a single displayed value (e.g., integer, floating point
    number, string).

    Design
    ----------
    All subclasses must support either (in order of descending preference):

    #. Explicit clipboard integration via the ``copy``, ``cut``, or ``paste``
       methods, typically supported by textual scalar widgets (e.g.,
       :class:`QLineEdit`).
    #. Implicit clipboard integration via the analogous Ctrl-c, Ctrl-x, and
       Ctrl-v keyboard shortcus, typically supported by numeric scalar widgets
       (e.g., :class:`QSpinBox`).

    If the current subclass does *not* define the ``copy``, ``cut``, and
    ``paste`` methods, this base class assumes this subclass to support the
    standard clipboard keyboard shortcuts instead. If this subclass supports
    neither, this base class silently reduces to a noop.
    '''

    # ..................{ SUPERCLASS ~ mandatory : method    }..................
    def copy_selection_to_clipboard(self) -> None:

        # If this subclass provides explicit clipboard support, prefer that.
        if objects.is_method(self, 'copy'):
            self.copy()
        # Else, assume (without evidence, of course) that this subclass supports
        # the standard keyboard shortcuts. In this case, queue up a key event
        # submitting the appropriate keyboard shortcut to this widget.
        else:
            QApplication.postEvent(self, QKeyEvent(
                QEvent.KeyPress, Qt.Key_C, Qt.ControlModifier))


    def cut_selection_to_clipboard(self) -> None:

        if objects.is_method(self, 'cut'):
            self.cut()
        else:
            QApplication.postEvent(self, QKeyEvent(
                QEvent.KeyPress, Qt.Key_X, Qt.ControlModifier))


    def paste_clipboard_to_selection(self) -> None:

        if objects.is_method(self, 'paste'):
            self.paste()
        else:
            QApplication.postEvent(self, QKeyEvent(
                QEvent.KeyPress, Qt.Key_V, Qt.ControlModifier))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
:mod:`PySide2`-based stack widget page controllers specific to tissue settings.
'''

# ....................{ IMPORTS                            }....................
# from PySide2.QtCore import QCoreApplication #, Signal, Slot
from PySide2.QtWidgets import QMainWindow
from betse.lib.yaml.abc.yamlabc import YamlABC
from betse.science.config.model.conftis import SimConfTissueDefault
from betse.util.type.obj import objects
# from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ SUBCLASSES                         }....................
#FIXME: Generalize to non-default tissue profiles.
class QBetseeSimConfTissueDefaultStackedWidgetPager(QBetseeControllerABC):
    '''
    :mod:`PySide2`-based stack widget page controller, connecting all editable
    widgets of the tissue page with the corresponding low-level settings of the
    current simulation configuration.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def init(self, main_window: QMainWindow) -> None:

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Simulation configuration state object.
        sim_conf = main_window.sim_conf

        # YAML-backed simulation subconfiguration whose class declares all
        # data descriptor-driven aliases referenced below.
        tissue_default = sim_conf.p.tissue_default

        # Initialize all general-purpose widgets on this page.
        main_window.sim_conf_tis_default_name.init(
            sim_conf=sim_conf,
            sim_conf_alias=SimConfTissueDefault.name,
            sim_conf_alias_parent=tissue_default,
        )

        # Initialize all filename-specific widgets on this page.
        main_window.sim_conf_tis_default_image_line.init(
            sim_conf=sim_conf,
            sim_conf_alias=SimConfTissueDefault.picker_image_filename,
            sim_conf_alias_parent=tissue_default,
            push_btn   =main_window.sim_conf_tis_default_image_btn,
            image_label=main_window.sim_conf_tis_default_image_label,
        )

        # Initialize all ion-specific widgets on this page.
        self._init_widgets_ion(
            main_window=main_window, page_conf=tissue_default)


    #FIXME: Can this be generalized to custom ion profiles as well?
    def _init_widgets_ion(
        self, main_window: QMainWindow, page_conf: YamlABC) -> None:
        '''
        Initialize all ion-specific widgets on this page.

        Attributes
        ----------
        main_window : QMainWindow
            Main window singleton.
        page_conf : YamlABC
            YAML-backed simulation subconfiguration specific to this page.
        '''

        #FIXME: Consider shifting into BETSE itself -- perhaps in a new
        #"betse.science.simulate.simion" submodule?

        # Set of the abbreviated names of all available ions.
        ION_NAMES = {'Na', 'K', 'Cl', 'Ca', 'M', 'P',}

        # Simulation configuration state object.
        sim_conf = main_window.sim_conf

        # For the abbreviated name of each supported ion...
        for ion_name in ION_NAMES:
            # Name of the instance variable of the main window providing the
            # widget on this page editing the default tissue profile's membrane
            # diffusion constant for this ion.
            ion_widget_name = 'sim_conf_tis_default_mem_' + ion_name

            # Name of the data descriptor of the default tissue profile's class
            # providing this membrane diffusion constant.
            ion_descriptor_name = 'Dm_' + ion_name

            # This widget.
            ion_widget = objects.get_attr(
                obj=main_window, attr_name=ion_widget_name)

            # This data descriptor.
            ion_descriptor = objects.get_attr(
                obj=SimConfTissueDefault, attr_name=ion_descriptor_name)

            # Initialize this widget with this data descriptor.
            ion_widget.init(
                sim_conf=sim_conf,
                sim_conf_alias=ion_descriptor,
                sim_conf_alias_parent=page_conf,
            )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
General-purpose :mod:`QLabel` subclasses.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import Qt, QCoreApplication, QSize  #, Signal, Slot
from PySide2.QtGui import QPixmap
from PySide2.QtWidgets import QFrame, QLabel, QScrollArea, QSizePolicy
from betse.lib.pil import pils
from betse.util.io.log import logs
from betse.util.path import pathnames, paths
from betse.util.type.cls import classes
from betse.util.type.text import mls
from betse.util.type.types import type_check
from betsee.util.path import guifiletype
from betsee.util.widget import guiwdg
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ SUBCLASSES                         }....................
#FIXME: Submit this class once working as a novel solution to:
#    https://stackoverflow.com/questions/8211982/qt-resizing-a-qlabel-containing-a-qpixmap-while-keeping-its-aspect-ratio
#
#In particular, note that the principal deficiency of the top-rated answer is
#its use of:
#
#    # This...
#    pixmap_size_new = self.size()
#
#    # ...rather than this.
#    pixmap_size_new = self.sizeHint()
#
#The former instructs Qt to gradually increase the size of this label to this
#preferred size over a lengthy (and hence unacceptable) period of ten seconds!

class QBetseeLabelImage(QBetseeObjectMixin, QLabel):
    '''
    :mod:`QLabel`-based widget preserving the aspect ratio of the optional
    **pixmap** (i.e., in-memory image) added to this widget.

    By default, pixmaps added to :class:`QLabel` widgets are stretched to fit
    layout requirements -- commonly resulting in unappealing deformation of
    pixmaps added to these widgets. This derivative widget corrects this harmful
    behaviour by constraining the aspect ratio of this widget to be equal to the
    aspect ratio of the contained pixmap if any. To do so:

    * The width of this widget is always preserved as is.
    * The height of this widget is dynamically set to be the width of this
      widget multiplied by the aspect ratio of the contained pixmap if any, thus
      constraining the aspect ratio of this widget to be the same.

    This widget supports all image filetypes supported by standard Qt widgets.
    Specifically, all images whose filetypes are in the system-specific set
    returned by the :func:`betse.util.path.guifiletype.get_image_read_filetypes`
    function are explicitly supported.

    Caveats
    ----------
    This widget *must* be contained in a :class:`QScrollArea` widget.
    Equivalently, some transitive parent widget of this widget *must* be a
    :class:`QScrollArea` widget. If this is *not* the case when the :meth:`init`
    method finalizing this widget's initialization is called, that method
    explicitly raises an exception.

    Why? Because size hints for widgets residing outside of :class:`QScrollArea`
    widgets are typically silently ignored. This widget declares a size hint
    preserving the aspect ratio of its pixmap. If this size hint is silently
    ignored, this pixmap's aspect ratio will typically be silently violated.
    Since this widget type exists *only* to preserve this aspect ratio, this
    constitutes a fatal error.

    Recursion
    ----------
    There exists circumstantial online evidence that this widget can
    *effectively* induces infinite recursion in edge cases. Specifically, `it
    has been asserted elsewhere <recursion claim_>`__ that:

    * :meth:`QBetseeLabelImage.resizeEvent` calls
      :meth:`QBetseeLabelImage.setPixmap`.
    * :meth:`QBetseeLabelImage.setPixmap` calls
      :meth:`QLabel.setPixmap`.
    * :meth:`QLabel.setPixmap` calls :meth:`QLabel.updateLabel`.
    * :meth:`QLabel.updateLabel` calls :meth:`QLabel.updateGeometry`.
    * :meth:`QLabel.updateGeometry` may conditionally call :meth:`QLabel.resize`
      in edge cases.
    * :meth:`QLabel.resize` queues a new resize event for this label.
    * The Qt event loop processes this event by calling
      :meth:`QLabel.resizeEvent`, completing the recursive cycle.

    Technically speaking, no recursive cycle exists. Due to indirection
    introduced by event handling, the :meth:`QLabel.resize` and
    :meth:`QLabel.resizeEvent` methods are called in different branches of the
    call stack. Ergo, the above call chain should noticeably degrade application
    performance *without* fatally exhausting the stack. Sadly, since these
    methods are called sequentially rather than recursively, detecting and
    guarding against this edge case is infeasible. (It could be worse.)

    Practically speaking, we were unable to replicate this worst-case issue.
    Since we cannot replicate it, we cannot resolve it. We have elected instead
    to do nothing, accepting that this may become a demonstrable issue later.

    .. _recursion claim:
        https://stackoverflow.com/a/41403419/2809027

    Attributes
    ----------
    _pixmap : (QPixmap, NoneType)
        Pixmap added to this widget by an external call to the :meth:`setPixmap`
        method if that method has been called *or* ``None`` otherwise.

    See Also
    ----------
    https://stackoverflow.com/a/22618496/2809027
        StackOverflow answer inspiring this implementation.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._pixmap = None

        # Lightly border this label's pixmap for aesthetics.
        self.setFrameShape(QFrame.StyledPanel)

        # Prevent Qt from auto-rescaling either this label's pixmap or text, as
        # Qt does so in a manner *NOT* preserving the aspect ratio of the
        # former. Although this boolean is already disabled by default, enforce
        # this constraint for safety by explicitly disabling this boolean.
        self.setScaledContents(False)


    def init(self) -> None:
        '''
        Finalize the initialization of this widget.

        Raises
        ----------
        BetseePySideWidgetException
            If this widget is *not* contained in a :class:`QScrollArea` widget.
            See the class docstring for commentary.
        '''

        # Finalize the initialization of our superclass.
        super().init()

        # If this widget is *NOT* in a scroll area, raise an exception.
        guiwdg.die_unless_widget_parent_satisfies(
            widget=self,
            predicate=lambda widget_parent: isinstance(
                widget_parent, QScrollArea))

    # ..................{ PROPERTIES                         }..................
    @property
    def _is_pixmap(self) -> bool:
        '''
        ``True`` only if this label has a pixmap that is *not* the null pixmap.
        '''

        return self._pixmap is not None and not self._pixmap.isNull()

    # ..................{ SUPERCLASS ~ getters               }..................
    def sizeHint(self) -> QSize:
        '''
        Preferred size for this label, defined in a manner preserving both this
        label's existing width *and* this label's pixmap's aspect ratio.

        If this property is *not* overridden in this manner, the
        :meth:`setPixmap` method rescaling this pixmap and hence this label
        gradually increase the size of this label to this preferred size over a
        lengthy (and hence unacceptable) period of nearly ten seconds. Hence,
        overriding this is non-optional.
        '''

        # Current width of this label.
        label_width = self.width()

        # Return this label's preferred size in a manner preserving both this
        # label's existing width *AND* this label's pixmap's aspect ratio.
        return QSize(label_width, self.heightForWidth(label_width))


    def heightForWidth(self, label_width_new: int) -> int:
        '''
        Preferred height for this widget given the passed width if this widget's
        height depend on its width *or* -1 otherwise (i.e., if this widget's
        height is independent of its width).

        Parameters
        ----------
        label_width_new : int
            Width that this widget is being externally resized to.

        Returns
        ----------
        int
            Either:
            * If this widget contains a **non-null pixmap** (i.e., if the
              :meth:`setPixmap` method of this widget has been passed a pixmap
              whose :meth:`QPixmap.isNull` method returns ``False``), the
              returned height is guaranteed to preserve the aspect ratio of this
              pixmap with respect to the passed width.
            * Else (i.e., if this widget either contains no pixmap *or* contains
              the null pixmap), the returned height is this widget's current
              height unmodified.
        '''

        # If this label has a non-null pixmap, return a height preserving the
        # aspect ratio of this pixmap with respect to this width, trivially
        # rounded from a float to integer.
        if self._is_pixmap:
            return round(
                (self._pixmap.height() * label_width_new) /
                 self._pixmap.width())
        # Else, this label either has no pixmap *OR* has the null pixmap. In
        # either case, return this label's current height as is.
        else:
            return self.height()

    # ..................{ SUPERCLASS ~ setters               }..................
    @type_check
    def setPixmap(self, pixmap: QPixmap) -> None:
        '''
        Set this label's pixmap to the passed pixmap, internally rescaled to
        this label's current size.

        Under ideal circumstances, this label's current size is identical to
        this label's **preferred size** (i.e., the size returned by the
        :meth:`sizeHint` method). Since the latter is overridden by this
        widget to preserve the aspect ratio of this pixmap, this pixmap's new:

        * Width will be the current width of this label.
        * Height will be a function of this width preserving the aspect ratio of
          this pixmap.

        For aesthetics, this pixmap is rescaled with smooth bilinear filtering
        rather than with the default non-smooth transition.

        Parameters
        ----------
        pixmap : QPixmap
            Pixmap to be set as this label's pixmap
        '''

        # Classify this pixmap and dimensions thereof for usability.
        self._pixmap = pixmap

        # Current size of this pixmap for logging purposes.
        pixmap_size_old = pixmap.size()

        # Size to resize pixmap to, equal to the preferred size of this label.
        # This size preserves the aspect ratio of this pixmap such that its:
        #
        # * Width is the current width of this label.
        # * Height is a function of this width preserving the aspect ratio of
        #   this pixmap.
        #
        # Ideally, the size() method returning the current size of this label
        # rather than the sizeHint() method returning only a preferred size
        # would be called. This widget has overriden all available size-specific
        # hint methods to return sizes preserving the aspect ratio of this
        # pixmap, but the size returned by the size() method still insists on
        # ignoring these hints in common edge cases.
        #
        # In particular, supplanting sizeHint() by size() here instructs Qt to
        # gradually increase the size of this label to this preferred size over
        # a lengthy (and hence unacceptable) period of nearly ten seconds.
        pixmap_size_new = self.sizeHint()

        # Log this rescaling.
        logs.log_debug(
            'Rescaling label pixmap from %dx%d to %dx%d...',
            pixmap_size_old.width(),
            pixmap_size_old.height(),
            pixmap_size_new.width(),
            pixmap_size_new.height())

        # Pixmap rescaled from the passed pixmap.
        pixmap_rescaled = pixmap.scaled(
            pixmap_size_new,  # * pixmap.devicePixelRatioF(),

            # Preserve the passed pixmap's existing aspect ratio. While the
            # passed size *SHOULD* already ensure this, we go the distance.
            Qt.KeepAspectRatio,

            # Rescale this pixmap with smooth bilinear filtering. By default,
            # pixmaps are rescaled abruptly rather than with this transition.
            Qt.SmoothTransformation,
        )
        # scaled.setDevicePixelRatio(devicePixelRatioF())

        # Set this label's pixmap to this rescaled pixmap.
        super().setPixmap(pixmap_rescaled)

    # ..................{ SUPERCLASS ~ events                }..................
    def resizeEvent(self, *args, **kwargs) -> None:

        # Defer to our superclass implementation.
        super().resizeEvent(*args, **kwargs)

        # If this label has a pixmap that is *NOT* the null pixmap...
        if self._is_pixmap:
            logs.log_debug('Rescaling pixmap after resizing label...')
            # Rescale this pixmap to the desired size of this widget.
            self.setPixmap(self._pixmap)

    # ..................{ LOADERS                            }..................
    @type_check
    def load_image(self, filename: str) -> None:
        '''
        Load the image with the passed filename as this label's pixmap in a
        sensible manner preserving the aspect ratio of this image.

        For safety, this method preferentially displays otherwise fatal errors
        resulting from this loading process as non-fatal warnings set as this
        label's text. Since this widget is typically *only* leveraged as an
        image previewer, the failure to preview arbitrary user-defined images of
        dubious origin, quality, and contents and possibly unsupported filetype
        should *not* halt the entire application. Ergo, it doesn't.

        Parameters
        ----------
        filename : str
            Absolute or relative filename of the image to be loaded.
        '''

        # Basename of this image's filename.
        basename = pathnames.get_basename(filename)

        # Filetype of this image if any *OR* "None" otherwise.
        filetype = pathnames.get_filetype_undotted_or_none(filename)

        # Log this preview attempt.
        logs.log_debug('Previewing image "%s"...', basename)

        # Attempt to...
        try:
            # If this image does *NOT* exist or does but is unreadable by the
            # current user, show a non-fatal warning and return.
            if not paths.is_readable(filename):
                self._warn(QCoreApplication.translate('QBetseeLabelImage',
                    'Image "{0}" not found or unreadable.'.format(filename)))
                return
            # Else, this image exists and is readable by the current user.

            # If this image has no filetype, show a warning and return.
            if filetype is None:
                self._warn(QCoreApplication.translate('QBetseeLabelImage',
                    'Image "{0}" has no filetype.'.format(filename)))
                return
            # Else, this image has a filetype.

            # Set of all image filetypes readable by Pillow.
            filetypes_pil = pils.get_filetypes()

            # Set of all image filetypes readable by Qt.
            filetypes_qt = guifiletype.get_image_read_filetypes()

            # If this image is unreadable by Pillow, show a warning and return.
            if filetype not in filetypes_pil:
                self._warn(QCoreApplication.translate('QBetseeLabelImage',
                    'Image filetype "{0}" unrecognized by Pillow.'.format(
                        filetype)))
                return
            # Else, this image is readable by Pillow and hence BETSE and is thus
            # presumably valid.

            # If this image is unreadable by Qt, show a warning and return.
            # Since Qt reads fewer image filetypes than Pillow, this condition
            # is effectively ignorable from the perspective of the end user.
            if filetype not in filetypes_qt:
                self._warn(QCoreApplication.translate('QBetseeLabelImage',
                    'Image filetype "{0}" not previewable.'.format(
                        filetype)))
                return
            # Else, this image is readable by both Pillow *AND* Qt.

            # In-memory pixmap trivially loaded from this on-disk image.
            pixmap = QPixmap(filename)

            # Set this label's pixmap as this pixmap.
            self.setPixmap(pixmap)

            # Resize this label to this message.
            # self.adjustSize()
        # If doing so raises an exception...
        except Exception as exception:
            # Exception type.
            exception_type = classes.get_class_name(exception)

            # Exception message.
            exception_message = str(exception)

            # Display this exception message as a non-fatal warning.
            self._warn(QCoreApplication.translate(
                'QBetseeLabelImage',
                'Image "{0}" preview failed with "{1}": {2}'.format(
                    basename,
                    exception_type,
                    exception_message)))

    # ..................{ WARNERS                            }..................
    @type_check
    def _warn(self, warning: str) -> None:
        '''
        Set the passed human-readable message as this label's text, log this
        message as a warning, and remove this label's existing pixmap if any.

        For clarity, this message is embedded in rich text (i.e., HTML) visually
        accentuating this message in a manner indicative of warnings -- notably,
        with boldened red text.

        Caveats
        ----------
        **This message is logged as is and hence assumed to be plaintext.** All
        HTML syntax (e.g., ``&amp;``, ``<table>``) embedded in this message is
        escaped for safety, preventing this message from being erroneously
        misinterpreted as rich text.

        Parameters
        ----------
        warning : str
            Human-readable warning message to be displayed as this label's text.
        '''

        # Log this warning.
        logs.log_warning(warning)

        # Remove this label's existing pixmap if any.
        self.clear()

        # Warning message with all embedded HTML syntax (e.g., "&amp;", "<b>")
        # escaped for safety.
        warning_escaped = mls.escape_ml(warning)

        # Set this label's text to rich text embedding this message in a
        # visually distinctive manner indicative of a warning.
        self._image_label.setText(QCoreApplication.translate(
            'QBetseeLabelImage',
            '<span style="color: #aa0000;"><b>Warning:</b> {0}</span>'.format(
                warning_escaped)))

        # Resize this label to this message.
        self.adjustSize()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application dependency** (i.e., both mandatory and optional
third-party Python packages required by this application) facilities.

This low-level submodule defines functions intended to be called by high-level
submodules (e.g., :mod:`betse.util.cli.cliabc`) *before* attempting to import any
such dependencies.
'''

#FIXME: Ideally eliminate the boilerplate repeated across the equivalent of
#this submodule in BETSE, BETSEE, and now NIMME by generalizing this
#functionality. To do so, see "nimme.lib.nimlib" for detailed commentary.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory dependencies,
# the top-level of this module may import *ONLY* from packages guaranteed to
# exist at initial runtime (i.e., standard Python and application packages,
# including both BETSEE and BETSE packages).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.lib import libs as betse_libs
from betse.util.io.log import logs
from betse.util.type.mapping import mappings
from betse.util.type.types import type_check
from betsee import guimetadata, guimetadeps

# ....................{ EXCEPTIONS                         }....................
def die_unless_runtime_mandatory_all() -> None:
    '''
    Raise an exception unless all mandatory runtime dependencies of this
    application are **satisfiable** (i.e., both importable and of a satisfactory
    version) *and* all external commands required by these dependencies (e.g.,
    GraphViz's ``dot`` command) reside in the current ``${PATH}``.

    Raises
    ----------
    BetseLibException
        If at least one mandatory runtime dependency is unsatisfiable.

    See Also
    ----------
    :func:`betse_libs.die_unless_runtime_mandatory_all`
        Further details.
    '''

    # Ideally, this function would reduce to the following one-liner:
    #    betse_libs.die_unless_requirements_dict(guimetadeps.RUNTIME_MANDATORY)
    #
    # Sadly, doing so invites subtle setuptools issues for developers and
    # contributers installing BETSE via "python3 setup.py develop". In that
    # case, the setuptools-specific version of BETSE is effectively frozen at
    # installation time. Since this version rapidly becomes desynchronized from
    # the physical version of BETSE in development (i.e., the
    # "betse.__version__" attribute), the setuptools-specific version of BETSE
    # must be ignored for validation purposes. Note, however, that the initial
    # "betsee.__main__" submodule has already guaranteed BETSE to both be
    # importable and of a sufficient version; ergo, re-validating BETSE here
    # would be entirely superfluous.
    #
    # Instead, this function removes BETSE from the set of all setuptools
    # requirements to be validated *BEFORE* validating those requirements.

    # Deep copy of these requirements excluding BETSE.
    RUNTIME_MANDATORY_SANS_BETSE = mappings.copy_map_sans_key(
        mapping=guimetadeps.RUNTIME_MANDATORY, key='BETSE')

    # Validate these requirements.
    betse_libs.die_unless_requirements_dict(RUNTIME_MANDATORY_SANS_BETSE)


@type_check
def die_unless_runtime_optional(*requirement_names: str) -> None:
    '''
    Raise an exception unless all optional runtime dependencies of this
    application with the passed :mod:`setuptools`-specific project names are
    **satisfiable** (i.e., both importable and of a satisfactory version)
    *and* all external commands required by these dependencies (e.g.,
    GraphViz's ``dot`` command) reside in the current ``${PATH}``.

    Parameters
    ----------
    requirement_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such name
        is unrecognized (i.e., is *not* a key of the
        :data:`guimetadeps.RUNTIME_OPTIONAL` dictionary), an exception is
        raised.

    Raises
    ----------
    BetseLibException
        If at least one such dependency is unsatisfiable.

    See Also
    ----------
    :func:`betse_libs.die_unless_runtime_mandatory_all`
        Further details.
    '''

    betse_libs.die_unless_requirements_dict_keys(
        guimetadeps.RUNTIME_OPTIONAL, *requirement_names)

# ....................{ TESTERS                            }....................
@type_check
def is_runtime_optional(*requirement_names: str) -> bool:
    '''
    ``True`` only if all optional runtime dependencies of this application with
    the passed :mod:`setuptools`-specific project names are **satisfiable**
    (i.e., both importable and of a satisfactory version) *and* all external
    commands required by these dependencies (e.g., GraphViz's ``dot`` command)
    reside in the current ``${PATH}``.

    Parameters
    ----------
    requirement_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such
        name is *not* a key of the :data:`guimetadeps.RUNTIME_OPTIONAL`
        dictionary and is thus unrecognized, an exception is raised.

    See Also
    ----------
    :func:`betse_libs.die_unless_runtime_mandatory_all`
        Further details.
    '''

    return betse_libs.is_requirements_dict_keys(
        guimetadeps.RUNTIME_OPTIONAL, *requirement_names)

# ....................{ IMPORTERS                          }....................
@type_check
def import_runtime_optional(*requirement_names: str) -> object:
    '''
    Import and return the top-level module object satisfying each optional
    runtime dependency of this application with the passed name.

    Parameters
    ----------
    requirement_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such name
        is unrecognized (i.e., is *not* a key of the
        :data:`guimetadeps.RUNTIME_OPTIONAL` dictionary), an exception is
        raised.

    See Also
    ----------
    :func:`betse_libs.import_requirements_dict_keys`
        Further details.
    '''

    return betse_libs.import_requirements_dict_keys(
        guimetadeps.RUNTIME_OPTIONAL, *requirement_names)

# ....................{ INITIALIZERS                       }....................
def reinit() -> None:
    '''
    (Re-)initialize all mandatory runtime dependencies of this application,
    thus including both BETSE and BETSEE.
    '''

    # Defer heavyweight imports.
    from betsee.util.app import guiapp

    # Log this initialization. Since initializing heavyweight third-party
    # dependencies (especially matplotlib) consumes non-trivial time, this
    # message is intentionally exposed to all users by default.
    logs.log_info('Loading third-party %s dependencies...', guimetadata.NAME)

    # Initialize PySide2 by instantiating the "QApplication" singleton *BEFORE*
    # initializing BETSE dependencies. Why? The reason is subtle, but critical:
    # initializing BETSE initializes matplotlib with the "Qt5Agg" backend, which
    # instantiates the "QApplication" singleton if this singleton has *NOT*
    # already been initialized. However, various application-wide settings
    # (e.g., metadata, high-DPI scaling emulation) *MUST* be initialized before
    # this singleton is instantiated. Permitting "Qt5Agg" to instantiate this
    # singleton first prevents us from initializing these settings here. This
    # singleton *MUST* thus be instantiated by us first.
    guiapp.init()

    # Initialize all mandatory runtime dependencies of BETSE.
    betse_libs.reinit(matplotlib_backend_name='Qt5Agg')



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
General-purpose :mod:`QScrollArea` subclasses.
'''

# ....................{ IMPORTS                           }....................
# from PySide2.QtCore import Signal, Slot
from PySide2.QtWidgets import QScrollArea
from betsee.util.widget.abc.guiwdgabc import QBetseeObjectMixin

# ....................{ SUBCLASSES                        }....................
class QBetseeScrollImage(QBetseeObjectMixin, QScrollArea):
    '''
    :mod:`QScrollArea`-based widget (pre)viewing a single image within a
    :mod:`QScrollArea`.

    This widget augments the stock :class:`QScrollArea` widget with:

    * **Image (pre)view.** This widget supports all image filetypes supported
      by stock Qt widgets. Specifically, all images whose filetypes are in the
      system-specific set returned by the
      :func:`betse.util.path.guifiletype.get_image_read_filetypes` function are
      explicitly supported.
    * **Horizontal scrollbars,** automatically displaying horizontal scrollbars
      for all columns whose content exceeds that column's width. For
      inexplicable reasons, this functionality has been seemingly intentionally
      omitted from the stock :class:`QScrollArea`.

    Attributes
    ----------
    _image_label : QLabel
        Child label contained with this line edit. To preview the image whose
        filename is the text displayed by this line edit, this label's pixmap is
        read from this filename. By convention, this label is typically situated
        below this line edit.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._image_label = None

    # ..................{ SUPERCLASS ~ setters              }..................
    def setColumnCount(self, column_count: int) -> None:

        # Defer to the superclass implementation.
        super().setColumnCount(column_count)

        # If this tree now contains more than one column, permit the last such
        # column's content to automatically resize to the viewport width.
        if column_count != 1:
            self.header().setStretchLastSection(True)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level application clipboard functionality.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QCoreApplication, QObject, Slot
from PySide2.QtWidgets import QWidget
from betsee.util.app import guiapp
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideClipboardException
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.util.io import guiclipboard
from betsee.util.io.key import guifocus
from betsee.util.type.guitype import QWidgetOrNoneTypes
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ CLASSES                            }....................
class QBetseeMainClipboard(QBetseeControllerABC):
    '''
    High-level **clipboarder** (i.e., :mod:`PySide2`-based object encapsulating
    all application clipboard state).

    This state includes:

    * Whether or not an application widget capable of receiving the interactive
      keyboard input focus is currently focused.
    * Whether or not the platform-specific system clipboard's plaintext buffer
      is currently empty (i.e., contains the empty string).

    Attributes
    ----------
    _widget_focused_if_any : QWidgetOrNoneTypes
        Widget currently receiving the interactive keyboard input focus if any
        *or* ``None`` otherwise. Ideally, the
        :func:`betsee.util.io.key.guifocus.get_widget_focused` getter would be
        called as needed to query for this widget rather than classifying this
        widget. Sadly, that getter appears to raise spurious exceptions on
        attempting to query for this widget when a toolbar button (notably, the
        copy, cut, or paste buttons) is clicked. Why? Presumably due to Qt
        subtleties in which the currently focused widget is temporarily
        defocused when a toolbar button is clicked -- regardless of that
        button's focus policy. In short, Qt issues.

    Attributes (Actions)
    ----------
    _action_copy : QAction
        Alias of the :attr:`QBetseeMainWindow.action_copy` action.
    _action_cut : QAction
        Alias of the :attr:`QBetseeMainWindow.action_cut` action.
    _action_paste : QAction
        Alias of the :attr:`QBetseeMainWindow.action_paste` action.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this clipboarder.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._action_copy  = None
        self._action_cut   = None
        self._action_paste = None
        self._widget_focused_if_any = None


    @type_check
    def init(self, main_window: QBetseeMainWindow) -> None:
        '''
        Initialize this clipboarder, owned by the passed main window widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of the system clipboard.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child widgets
        (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        '''

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Log this initialization.
        logs.log_debug('Sanitizing system clipboard state...')

        # Classify actions subsequently required by this object. Since this main
        # window owns this object, since weak references are unsafe in a
        # multi-threaded GUI context, and since circular references are bad,
        # this object does *NOT* retain a reference to this main window.
        self._action_copy  = main_window.action_copy
        self._action_cut   = main_window.action_cut
        self._action_paste = main_window.action_paste

        # Connect each such action to this object's corresponding slot.
        self._action_copy.triggered.connect(
            self._copy_widget_focused_selection_to_clipboard)
        self._action_cut.triggered.connect(
            self._cut_widget_focused_selection_to_clipboard)
        self._action_paste.triggered.connect(
            self._paste_clipboard_to_widget_focused_selection)
        # logs.log_debug('Wiring everything up...')

        # Application singleton, localized to avoid retaining references..
        gui_app = guiapp.get_app()

        # System clipboard.
        clipboard = guiclipboard.get_clipboard()

        # Connect all relevant application-wide slots to corresponding signals
        # on this main window. Since this application strictly adheres to the
        # SDI metaphor, there exists a one-to-one correspondence between this
        # application and this object. (That is, this application always
        # contains exactly one object of this type.)
        gui_app.focusChanged.connect(self._widget_focus_set)
        clipboard.dataChanged.connect(self._clipboard_text_set)

        # Set the state of all widgets dependent upon this simulation
        # configuration state *AFTER* connecting all relavant signals and slots.
        # Initially, no widget has yet to be focused.
        self._widget_focus_set(None, None)

    # ..................{ SLOTS ~ state                      }..................
    #FIXME: Is this slot's macOS-specific caveat (see docstring below) actually
    #an issue? Only if Qt actually ignores rather than defers clipboard changes
    #that occur when this application is *NOT* the active application. In that
    #case, we'll need to also define a new slot connecting to the application
    #activation signal (...whatever that is) manually invoking this slot.

    @Slot()
    def _clipboard_text_set(self) -> None:
        '''
        Slot signalled when any text is copied or cut into the system
        clipboard's plaintext buffer by any application in the current windowing
        session (including the current application).

        Caveats
        ----------
        The slot is signalled in a somewhat deferred manner on macOS. To quote
        the documentation for the :attr:`QClipboard.dataChanged` signal:

             On macOS and with Qt version 4.3 or higher, clipboard changes made
             by other applications will only be detected when the application is
             activated.
        '''

        # Currently focused widget if any or None otherwise.
        widget_focused_if_any = guifocus.get_widget_focused_or_none()

        # Update the state of all clipboard-dependent actions with this widget.
        self._widget_focus_set(widget_focused_if_any, widget_focused_if_any)


    @Slot(QWidget, QWidget)
    def _widget_focus_set(
        self,
        widget_focused_old: QWidgetOrNoneTypes,
        widget_focused_new: QWidgetOrNoneTypes,
    ) -> None:
        '''
        Slot signalled when an application widget loses and/or gains interactive
        keyboard input focus (e.g., due to the tab-key being pressed, this
        widget being clicked, or this main window being made active).

        The slot is signalled *after* both widgets have been notified of this
        :class:`QFocusEvent`.

        Parameters
        ----------
        widget_focused_old : QWidgetOrNoneTypes
            Previously focused widget if any *or* ``None`` otherwise.
        widget_focused_new : QWidgetOrNoneTypes
            Previously focused widget if any *or* ``None`` otherwise.
        '''

        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # WARNING: This is a fundamentally fragile slot. Exceptions accidentally
        # raised by this slot's implementation may induce infinite recursion.
        # See the "except" block below for further commentary.
        #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

        # Attempt to...
        try:
            # Log this focus change.
            # from betsee.util.widget.abc import guiwdgabc
            # logs.log_debug(
            #     'Changing focus from %s to %s...',
            #     guiwdgabc.get_label(widget_focused_old),
            #     guiwdgabc.get_label(widget_focused_new))

            # Classify this widget for subsequent lookup *BEFORE* calling the
            # _is_widget_clipboardable() method.
            self._widget_focused_if_any = widget_focused_new

            # True only if the currently focused widget is an
            # application-specific editable widget transparently supporting
            # copying, cutting, and pasting into and from the system clipboard.
            is_widget_focused_clipboardable = (
                self._is_widget_focused_clipboardable())

            # Enable or disable actions requiring such a widget to be focused.
            self._action_copy.setEnabled(is_widget_focused_clipboardable)
            self._action_cut .setEnabled(is_widget_focused_clipboardable)

            # Enable or disable actions requiring such a widget to be focused
            # *AND* the system clipboard's plaintext buffer to be non-empty.
            #
            # This overly simplistic logic would typically invite
            # desynchronization issues. Namely, if this buffer is currently
            # empty but subsequently copied or cut into with a non-empty string
            # by any application (including the current application), the state
            # of this action would be desynchronized from the state of this
            # buffer. To circumvent these issues, the _clipboard_text_set() slot
            # defers to this slot on each change to the state of this buffer and
            # hence guarantees this logic to be reevaluated as needed.
            self._action_paste.setEnabled(
                is_widget_focused_clipboardable and
                guiclipboard.is_clipboard_text())
        # If an exception is raised, avoid infinite recursion in the Qt event
        # loop by permanently disconnecting this slot from its corresponding
        # signal *BEFORE* this exception is propagated up the callstack. While
        # slightly destructive, this is the least-worst option.
        #
        # Failing to do so provokes the following infinite recursion:
        #
        # * This slot raises an exception.
        # * This exception is propagated up to the default exception handler.
        # * This handler displays a PySide2 widget graphically presenting this
        #   exception to the user.
        # * This widget implicitly obtains the interactive keyboard input focus.
        # * This focus change invokes the signal connected to this slot.
        # * This slot raises an exception.
        #
        # The only alternatives would be to:
        #
        # * Prevent the PySide2 widget displayed by the default exception
        #   handler from obtaining the focus -- a fragile, platform-specific,
        #   and possibly unenforceable constraint in the best case.
        # * Call the qApp.blockSignals() method, preventing the "QApplication"
        #   singleton from signalling *ANY* other slots -- which is even more
        #   heavy-handed and hence undesirable than the current approach.
        except:
            # Application singleton, localized to avoid retaining references..
            gui_app = guiapp.get_app()

            # Disconnect this signal from this slot... *PERMANENTLY.*
            gui_app.focusChanged.disconnect(self._widget_focus_set)

            # Propagate this exception up the callstack.
            raise

    # ..................{ SLOTS ~ action                     }..................
    @Slot()
    def _copy_widget_focused_selection_to_clipboard(self) -> None:
        '''
        Slot invoked in response to a user-driven request to copy the
        currently focused widget's **current selection** (i.e., currently
        selected subset of this widget's value(s)) to the system clipboard,
        silently replacing the prior contents if any.
        '''

        # If no clipboardable widget is currently focused, raise an exception.
        self._die_unless_widget_focused_clipboardable()

        # Log this copy.
        logs.log_debug(
            'Copying widget "%s" selection to clipboard...',
            self._widget_focused_if_any.obj_name)

        # Copy this widget's current selection to the clipboard.
        self._widget_focused_if_any.copy_selection_to_clipboard()


    @Slot()
    def _cut_widget_focused_selection_to_clipboard(self) -> None:
        '''
        Slot invoked in response to a user-driven request to **cut** (i.e., copy
        and then remove as a single atomic operation) the currently focused
        widget's **current selection** (i.e., currently selected subset of this
        widget's value(s)) to the clipboard's plaintext buffer, silently
        replacing the prior contents if any.
        '''

        # If no clipboardable widget is currently focused, raise an exception.
        self._die_unless_widget_focused_clipboardable()

        # Log this cut.
        logs.log_debug(
            'Cutting widget "%s" selection to clipboard...',
            self._widget_focused_if_any.obj_name)

        # Cut this widget's current selection to the clipboard.
        self._widget_focused_if_any.cut_selection_to_clipboard()


    @Slot()
    def _paste_clipboard_to_widget_focused_selection(self) -> None:
        '''
        Slot invoked in response to a user-driven request to paste the contents
        of the system clipboard over the currently focused widget's **current
        selection** (i.e., currently selected subset of this widget's value(s)),
        silently replacing the prior selection if any.
        '''

        # If no clipboardable widget is currently focused, raise an exception.
        self._die_unless_widget_focused_clipboardable()

        # Log this paste.
        logs.log_debug(
            'Pasting clipboard over widget "%s" selection...',
            self._widget_focused_if_any.obj_name)

        # Paste the clipboard over this widget's current selection.
        self._widget_focused_if_any.paste_clipboard_to_selection()

    # ..................{ EXCEPTIONS                         }..................
    def _die_unless_widget_focused_clipboardable(self) -> None:
        '''
        Raise an exception unless an application-specific widget transparently
        supporting copying, cutting, and pasting into and from the system
        clipboard is currently focused.

        Raises
        ----------
        BetseePySideFocusException
            If *no* widget is currently focused.
        BetseePySideClipboardException
            If a widget is currently focused but this widget is *not* an
            application-specific editable widget supporting clipboard operation.
        '''

        # If this widget is *NOT* an application-specific editable widget, raise
        # an exception.
        if not self._is_widget_focused_clipboardable():
            raise BetseePySideClipboardException(QCoreApplication.translate(
                'QBetseeMainClipboard',
                'Focused widget "{0}" not clipboardable (i.e., not '
                'an instance of "QBetseeEditWidgetMixin" '
                'whose "is_clipboardable" property is True).'.format(
                    self._widget_focused_if_any)))

    # ..................{ TESTERS                            }..................
    @type_check
    def _is_widget_focused_clipboardable(self) -> bool:
        '''
        ``True`` only if the currently focused widget (if any) is an
        application-specific editable widget transparently supporting copying,
        cutting, and pasting into and from the system clipboard.

        Returns
        ----------
        bool
            ``True`` only if this widget is an application-specific
            clipboardable widget.
        '''

        # Avoid circular import dependencies.
        from betsee.util.widget.abc.guiclipboardabc import (
            QBetseeClipboardWidgetMixin)

        # Return True if the currently focused widget if any is clipboardable.
        return isinstance(
            self._widget_focused_if_any, QBetseeClipboardWidgetMixin)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Abstract base classes of most application-specific widget subclasses.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific submodules of this subpackage (i.e.,
# "betsee.util.widget"). Since those submodules must *ALWAYS* be able to safely
# import from this submodule, circularities are best avoided here.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from PySide2.QtWidgets import QUndoStack
from betse.util.io.log import logs
from betse.util.py import pyident
from betse.util.type.cls import classes
from betse.util.type.text import strs
from betse.util.type.types import type_check
from betsee.guiexception import BetseePySideEditWidgetException

# ....................{ GLOBALS                            }....................
_OBJ_NAME_DEFAULT = 'N/A'
'''
Default Qt object name for all :class:`QBetseeObjectMixin` instances.
'''

# ....................{ MIXINS                             }....................
# To avoid metaclass conflicts with the "QWidget" base class inherited by all
# widgets also inheriting this base class, this base class *CANNOT* be
# associated with another metaclass (e.g., "abc.ABCMeta").
class QBetseeObjectMixin(object):
    '''
    Abstract base class of most application-specific Qt object subclasses.

    This class is suitable for use as a multiple-inheritance mixin. To preserve
    the expected method resolution order (MRO) semantics, this class should
    typically be inherited *first* rather than *last* in subclasses.

    Attributes (Private)
    ----------
    _is_initted : bool
        ``True`` only if this object's :meth:`init` method has been called.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this application-specific Qt object.

        Parameters
        ----------
        All parameters are passed as is to the superclass this mixin is mixed
        into (e.g., :class:`QObject` or a subclass thereof).

        Caveats
        ----------
        **Subclasses overriding this method should not attempt to accept
        subclass-specific parameters.** Due to the semantics of Python's
        method-resolution order (MRO), accidentally violating this constraint is
        guaranteed to raise non-human-readable exceptions at subclass
        instantiation time.

        Abstract base subclasses may trivially circumvent this constraint by
        defining abstract properties which concrete subclasses then define. When
        doing so, note that abstract methods should raise the
        :class:`BetseMethodUnimplementedException` exception rather than be
        decorated by the usual :meth:`abstractmethod` decorator -- which is
        *not* safely applicable to subclasses of this class.

        For example:

            >>> from betse.exceptions import BetseMethodUnimplementedException
            >>> @property
            ... def muh_subclass_property(self) -> MuhValueType:
            ...     raise BetseMethodUnimplementedException()
        '''

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Nullify all remaining instance variables for safety.
        self._is_initted = False

        # If this object has no name, default this object's name.
        if not self.obj_name:
            self.obj_name = _OBJ_NAME_DEFAULT


    def init(self) -> None:
        '''
        Finalize the initialization of this Qt object.

        This method is principally intended to simplify the implementation of
        subclasses overriding this method with subclass-specific finalization.

        Raises
        ----------
        BetseePySideEditWidgetException
            If this method has already been called for this object, preventing
            objects from being erroneously refinalized.
        '''

        # If this object has a meaningful name, log this initialization; else,
        # the subclass is expected to log this initialization.
        if self.obj_name != _OBJ_NAME_DEFAULT:
            logs.log_debug('Initializing object "%s"...', self.obj_name)

        # If this object's initialization has already been finalized, raise an
        # exception.
        if self._is_initted:
            raise BetseePySideEditWidgetException(
                'Object "{}" already initialized.'.format(self.obj_name))

        # Record this object's initialization to now have been finalized.
        self._is_initted = True


    def init_if_needed(self, *args, **kwargs) -> None:
        '''
        Finalize the initialization of this object if needed (i.e., if this
        object's initialization has *not* already been finalized by a call to
        the :meth:`init` method).

        This method safely wraps the :meth:`init` method, effectively squelching
        the exception raised by that method when this object's initialization
        has already been finalized.

        Parameters
        ----------
        All parameters are passed as is to the :meth:`init` method if called.
        '''

        # If this object's initialization has *NOT* been finalized, do so.
        if self._is_initted:
            self.init(*args, **kwargs)

    # ..................{ PROPERTIES ~ read-only             }..................
    @property
    def is_initted(self) -> bool:
        '''
        ``True`` only if this object's :meth:`init` method has been called.
        '''

        return self._is_initted

    # ..................{ PROPERTIES                         }..................
    @property
    def obj_name(self) -> str:
        '''
        Qt-specific name of this object.

        This property getter is a convenience alias of the non-Pythonic
        :meth:`objectName` method.
        '''

        return self.objectName()


    @obj_name.setter
    @type_check
    def obj_name(self, obj_name: str) -> None:
        '''
        Set the Qt-specific name of this object to the passed string.

        This property setter is a convenience alias of the non-Pythonic
        :meth:`setObjectName` method.
        '''

        self.setObjectName(obj_name)

    # ..................{ SETTERS                            }..................
    def set_obj_name_from_class_name(self) -> None:
        '''
        Set the Qt-specific name of this object to the unqualified name of this
        subclass, altered to comply with object name standards (e.g., from
        ``QBetseeSimmerWorkerSeed`` to ``simmer_worker_seed``).

        Specifically, this function (in order):

        #. Obtains the unqualified name of this subclass.
        #. Removes any of the following prefixes from this name:

           * ``QBetsee``, the string prefixing the names of all
             application-specific :class:`QObject` subclasses.
           * ``Q``, the string prefixing the names of all
             application-agnostic :class:`QObject` subclasses.

        #. Converts this name from CamelCase to snake_case.
        #. Sets this object's name to this name.

        Design
        ----------
        This method is intentionally *not* called by the :meth:`__init__` method
        to set this object's name to a (seemingly) sane default. Why? Because
        numerous subclasses prefer to manually set this name. Unconditionally
        calling this method for every subclass would have the undesirable side
        effect of preventing this and other subclasses from detecting when the
        object name has yet to be set (e.g., via a comparison against the
        :data:`_OBJ_NAME_DEFAULT` default).
        '''

        # Subclass of this object.
        cls = type(self)

        # Unqualified CamelCase name of this subclass.
        cls_name = classes.get_name(cls)

        # Remove the application-specific class name prefix if found and then
        # remove the application-agnostic class name prefix if found, as the
        # latter is a substring of the former.
        cls_name = strs.remove_prefix_if_found(text=cls_name, prefix='QBetsee')
        cls_name = strs.remove_prefix_if_found(text=cls_name, prefix='Q')

        # Set this object's name to this CamelCase converted into snake_case.
        self.obj_name = pyident.convert_camelcase_to_snakecase(cls_name)

# ....................{ MIXINS ~ edit                      }....................
class QBetseeEditWidgetMixin(QBetseeObjectMixin):
    '''
    Abstract base class of most application-specific **editable widget** (i.e.,
    widget interactively editing one or more values in an undoable manner)
    subclasses.

    Attributes
    ----------
    is_undo_cmd_pushable : bool
        ``True`` only if undo commands are safely pushable from this widget onto
        the undo stack *or* ``False`` when either:
        * This widget's content is currently being programmatically populated.
        * A previous undo command is already being applied to this widget.
        In both cases, changes to this widget's content are program- rather than
        user-driven and hence are *NOT* safely undoable. If ``False``, widget
        subclass slots intending to push an undo commands onto the undo stack
        should instead (in order):
        * Temporarily avoid doing so for the duration of the current slot call,
          as doing so *could* induce infinite recursion.
        * Set ``self.is_undo_cmd_pushable = False`` to permit all subsequent
          slot calls to push undo commands onto the undo stack.
    _undo_stack : QUndoStack
        Undo stack to which this widget pushes undo commands if any *or*
        ``None`` otherwise.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed arguments.
        super().__init__(*args, **kwargs)

        # Unset the undo stack to which this widget pushes undo commands.
        self._unset_undo_stack()

    # ..................{ UNDO STACK ~ set                   }..................
    @type_check
    def _set_undo_stack(self, undo_stack: QUndoStack) -> None:
        '''
        Set the undo stack to which this widget pushes undo commands, permitting
        the :meth:`_push_undo_cmd_if_safe` method to pushing undo commands from
        this widget onto this stack.
        '''

        # Classify all passed parameters.
        self._undo_stack = undo_stack

        # Undo commands are now safely pushable from this widget.
        self.is_undo_cmd_pushable = True


    def _unset_undo_stack(self) -> None:
        '''
        Unset the undo stack to which this widget pushes undo commands,
        preventing the :meth:`_push_undo_cmd_if_safe` method from pushing undo
        commands from this widget.
        '''

        # Classify all passed parameters.
        self._undo_stack = None

        # Undo commands are now safely pushable from this widget.
        self.is_undo_cmd_pushable = False

    # ..................{ UNDO STACK ~ other                 }..................
    def _is_undo_stack_dirty(self) -> bool:
        '''
        ``True`` only if the undo stack associated with this widget is in the
        **dirty state** (i.e., contains at least one undo command to be undone).
        '''

        return not self._undo_stack.isClean()


    @type_check
    def _push_undo_cmd_if_safe(
        self,
        # To avoid circular imports, this type is validated dynamically.
        undo_cmd: (
            'betsee.util.widget.abc.guiundocmdabc.QBetseeWidgetUndoCommandABC'),
    ) -> None:
        '''
        Push the passed widget-specific undo command onto the undo stack
        associated with this widget.

        Parameters
        ----------
        undo_cmd : QBetseeWidgetUndoCommandABC
            Widget-specific undo command to be pushed onto this stack.
        '''

        # If undo commands are *NOT* safely pushable (e.g., due to a prior undo
        # command currently being applied to this widget), silently noop.
        # Pushing this undo command unsafely could provoke infinite recursion.
        if not self.is_undo_cmd_pushable:
            return

        # Else, no such command is being applied. Since pushing this undo
        # command onto the stack is thus safe, do so.
        self._sim_conf.undo_stack.push(undo_cmd)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017 by Alexis Pietak & Cecil Curry
# See "LICENSE" for further details.

'''
Main entry point of this application's command line interface (CLI).

This submodule is a thin wrapper intended to be:

* Indirectly imported and run from external entry point scripts installed by
  setuptools (e.g., the ``betsee`` command).
* Directly imported and run from the command line (e.g., via
  ``python -m betsee.cli``).
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory
# dependencies, this module may import *ONLY* from standard Python packages and
# application-specific packages importing *ONLY* from standard Python packages.
# By definition, this excludes all third-party packages and most
# application-specific packages.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

import sys
from betsee import guimetadata
from betsee.guimetadata import NAME
from betsee.guimetadeps import BETSE_VERSION_REQUIRED_MIN

# ....................{ EXCEPTIONS                        }....................
class _BetseNotFoundException(Exception):
    '''
    Exception raised on detecting BETSE (i.e., this application's core
    mandatory dependency) to be unimportable by the active Python interpreter.

    Design
    ----------
    This subclass intentionally complies with the public API of the
    :class:`betsee.guiexception.BetseeException` superclass to enable duck
    typing between the two classes (e.g., by the
    :func:`betsee.util.io.ioerr.show_exception` function). As the
    :mod:`betsee.guiexception` submodule necessarily imports from and hence
    assumes the importability of the mandatory third-party :mod:`PySide2`
    package whose importability has *not* yet been validated at this early time
    in the application lifecycle, this submodule *cannot* safely import from
    that submodule and hence explicitly subclass that superclass.
    '''

    # ..................{ INITIALIZERS                      }..................
    def __init__(self, title: str, synopsis: str, exegesis: str) -> None:

        # Initialize our superclass.
        super().__init__(synopsis)

        # Classify all passed parameters to enable duck typing.
        self.title = title
        self.synopsis = synopsis
        self.exegesis = exegesis

# ....................{ MAIN                              }....................
def main(arg_list: list = None) -> int:
    '''
    Run this application's command-line interface (CLI) with the passed
    arguments if non-``None`` *or* with the arguments passed on the command
    line (i.e., :attr:`sys.argv`) otherwise.

    This function is provided as a convenience to callers requiring procedural
    functions rather than conventional methods (e.g., :mod:`setuptools`).

    Parameters
    ----------
    arg_list : list
        List of zero or more arguments to pass to this interface. Defaults to
        ``None``, in which case arguments passed on the command line (i.e.,
        :attr:`sys.argv`) will be used instead.

    Returns
    ----------
    int
        Exit status of this interface and hence this process as an unsigned
        byte (i.e., integer in the range ``[0, 255]``).
    '''

    # Validate BETSE to be satisfied *BEFORE* attempting to import from this
    # application's package tree, most of which requires BETSE.
    try:
        _die_unless_betse()
    # If BETSE is unsatisfied, display this exception in an appropriate manner
    # and return the exit status of doing so.
    except _BetseNotFoundException as exception:
        return _show_betse_exception(exception)

    # Import from this application's package tree *AFTER* validating BETSE.
    from betsee.cli.guicli import BetseeCLI

    # Run this application's CLI and return the exit status of doing so.
    return BetseeCLI().run(arg_list)

# ....................{ EXCEPTIONS                        }....................
def _die_unless_betse() -> None:
    '''
    Raise an exception unless BETSE, the principal mandatory dependency of this
    application, is **satisfied** (i.e., both importable and of a version
    greater than or equal to that required by this application).

    Raises
    ----------
    _BetseNotFoundException
        If BETSE is unsatisfied (i.e., either unimportable or of a version
        less than that required by this application).
    '''

    # Title of all exceptions raised below.
    EXCEPTION_TITLE = 'BETSE Unsatisfied'

    # Attempt to import BETSE.
    try:
        import betse
    # If BETSE is unimportable, chain this low-level "ImportError" exception
    # into a higher-level application-specific exception.
    except ImportError as import_error:
        raise _BetseNotFoundException(
            title=EXCEPTION_TITLE,
            synopsis='Mandatory dependency BETSE not found.',
            exegesis='Python package "betse" unimportable.',
        ) from import_error
    # Else, BETSE is importable.

    # Minimum version of BETSE required by this application as a
    # machine-readable tuple of integers. This tuple is only required once
    # (i.e., here) and hence *NOT* persisted as a "guimetadata" global.
    BETSE_VERSION_REQUIRED_MIN_PARTS = (
        guimetadata._convert_version_str_to_tuple(BETSE_VERSION_REQUIRED_MIN))

    # If the current version of BETSE is insufficient, raise an exception.
    if betse.__version_info__ < BETSE_VERSION_REQUIRED_MIN_PARTS:
        raise _BetseNotFoundException(
            title=EXCEPTION_TITLE,
            synopsis='Obsolete version of mandatory dependency BETSE found.',
            exegesis=(
                '{} requires at least BETSE {}, '
                'but only BETSE {} is currently installed.'.format(
                    NAME, BETSE_VERSION_REQUIRED_MIN, betse.__version__)))

    # Purely for testing purposes.
    # raise _BetseNotFoundException(
    #     title=EXCEPTION_TITLE,
    #     synopsis='BETSE not found.',
    #     exegesis='Python package "betse" not importable.',
    # )

# ....................{ EXCEPTIONS                        }....................
def _show_betse_exception(exception: _BetseNotFoundException) -> int:
    '''
    Display the passed exception signifying BETSE to be unsatisfied in an
    appropriate manner.

    Parameters
    ----------
    exception : _BetseNotFoundException
        Exception to be displayed.

    Returns
    ----------
    int
        Exit status implying failure (i.e., 1).
    '''

    assert isinstance(exception, _BetseNotFoundException), (
        '"{}" not a BETSE-not-found exception.'.format(exception))

    # Always redirect this exception message to the standard error file handle
    # for the terminal running this CLI command (if any).
    print(str(exception), file=sys.stderr)

    # Additionally attempt to...
    try:
        # Import PySide2.
        from betsee.util.io import guierr

        # Display a PySide2-based message box displaying this exception.
        guierr.show_exception(exception)
    # If PySide2 or any other module indirectly imported above is unimportable,
    # print this exception message in the same manner as above but otherwise
    # ignore this exception. Why? Because we have more significant fish to fry.
    except ImportError as import_error:
        print(str(import_error), file=sys.stderr)

    # Report failure to our parent process.
    return 1

# ....................{ MAIN                              }....................
# If this module is imported from the command line, run this application's CLI;
# else, noop. For POSIX compliance, the exit status returned by this function
# is propagated to the caller as this script's exit status.
if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation configurator** (i.e., :mod:`PySide2`-based object
both displaying *and* modifying external YAML-formatted simulation
configuration files) functionality.
'''

#FIXME: Either this or a subordinate "QObject" should routinely detect
#simulation configuration desynchronization (i.e., external changes to the
#on-disk YAML file underlying the currently open simulation configuration if
#any) and interactively query the current user to either:
#
#* Discard all in-GUI changes to this configuration and reload this
#  configuration from disk.
#* Preserve all in-GUI changes to this configuration by ignoring these external
#  changes, thus overwriting these changes on the next save.
#
#To do so, we'll need to track the current mtime of this file with a
#"QFileSystemWatcher" instance as follows:
#
#* Define a new resync_file() slot of this class accepting a single "str"
#  argument: the absolute filename of the watched file that was modified. In
#  this case, that will be the top-level YAML file. Naturally, this slot will
#  perform the bulk of the logic pertaining to resynchronization. Note that
#  we'll need to explicitly test whether this file still exists or not in this
#  slot, as it may have been moved away or removed entirely.
#* In the __init__() and/or init() methods of this class:
#  * Create a new "_path_watcher" variable of type "QFileSystemWatcher".
#  * Connect the fileChanged() signal of this variable.
#* In the set_filename() slot:
#  * When a new simulation configuration is opened (i.e., "filename" is
#    non-empty), call the "self._path_watcher.addPath(filename)" method.
#  * When an open simulation configuration is closed (i.e., "filename" is
#    empty), call the "self._path_watcher.removePath(filename)" method.
#
#Note also that files can be moved and removed from disk. If the
#on-disk YAML file underlying the currently open simulation configuration no
#longer exists at any point (e.g., due to being externally removed), what is the
#sanest response? Should we simply ignore this case? Doing so will simply
#silently rewrite this file to disk on the next save, which seems reasonably
#sane. In other words, the noop-based lazy approach may be the correct approach.

# ....................{ IMPORTS                           }....................
from PySide2.QtCore import QCoreApplication, Signal, Slot
from PySide2.QtWidgets import QMessageBox
from betse.science.config import confio
from betse.science.parameters import Parameters
from betse.util.io.log import logs
from betse.util.type.types import type_check, StrOrNoneTypes
from betsee.guiexception import BetseeSimConfException
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.util.app import guiappstatus
from betsee.util.io import guimessage
from betsee.util.path import guifile
from betsee.util.widget.abc.guicontrolabc import QBetseeControllerABC

# ....................{ CLASSES                           }....................
class QBetseeSimConf(QBetseeControllerABC):
    '''
    High-level **simulation configurator** (i.e., :mod:`PySide2`-based object
    both displaying *and* modifying external YAML-formatted simulation
    configuration files).

    This configurator maintains all state required to manage these files,
    including:

    * Whether or not a simulation configuration is currently open.
    * Whether or not an open simulation configuration has unsaved changes.

    Attributes (Public)
    ----------
    p : Parameters
        High-level simulation configuration encapsulating a low-level
        dictionary parsed from an even lower-level YAML-formatted file. By
        design, this object is guaranteed to be a **singleton** (i.e., remain
        the same object for the lifetime of this application).
    undo_stack : QBetseeUndoStackSimConf
        Undo stack for the currently open simulation configuration if any *or*
        the empty undo stack otherwise.

    Attributes (Private: Non-widgets)
    ----------
    _is_dirty : bool
        ``True`` only if a simulation configuration is currently open *and*
        this configuration is **dirty** (i.e., has unsaved changes).

    Attributes (Private: Widgets)
    ----------
    _action_make_sim : QAction
        Alias of the :attr:`QBetseeMainWindow.action_make_sim` action.
    _action_open_sim : QAction
        Alias of the :attr:`QBetseeMainWindow.action_open_sim` action.
    _action_close_sim : QAction
        Alias of the :attr:`QBetseeMainWindow.action_close_sim` action.
    _action_save_sim : QAction
        Alias of the :attr:`QBetseeMainWindow.action_save_sim` action.
    _action_save_sim_as : QAction
        Alias of the :attr:`QBetseeMainWindow.action_save_sim_as` action.
    _sim_conf_changed_signal : QSignal
        Alias of the :attr:`QBetseeMainWindow.sim_conf_changed_signal` signal.
    _sim_conf_stack : QBetseeSimConfStackedWidget
        Alias of the :attr:`QBetseeMainWindow.sim_conf_stack` widget.
    _sim_conf_tree : QBetseeSimConfTreeWidget
        Alias of the :attr:`QBetseeMainWindow.sim_conf_tree` widget.
    _sim_conf_tree_frame : QFrame
        Alias of the :attr:`QBetseeMainWindow.sim_conf_tree_frame` widget.
    _sim_tab : QBetseeSimmerTabWidget
        Alias of the :attr:`QBetseeMainWindow.sim_tab` widget.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, *args, **kwargs) -> None:
        '''
        Initialize this configurator.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Nullify all instance variables for safety.
        self._is_dirty = False
        self._action_make_sim = None
        self._action_open_sim = None
        self._action_close_sim = None
        self._action_save_sim = None
        self._action_save_sim_as = None
        self._sim_conf_stack = None
        self._sim_conf_tree = None
        self._sim_conf_tree_frame = None
        self._sim_tab = None
        self.undo_stack = None

        # High-level simulation configuration, defaulting to the unload state.
        self.p = Parameters()


    @type_check
    def init(self, main_window: QBetseeMainWindow) -> None:
        '''
        Finalize this configurator's initialization, owned by the passed main
        window widget.

        This method connects all relevant signals and slots of *all* widgets
        (including the main window, top-level widgets of that window, and leaf
        widgets distributed throughout this application) whose internal state
        pertains to the high-level state of this simulation configuration.

        Specifically:

        * When any such widget's content (e.g., editable text, selectable item)
          is modified, widgets elsewhere are notified of this simulation
          configuration modification via the
          :attr:`set_dirty_signal` signal.

        To avoid circular references, this method is guaranteed to *not* retain
        references to this main window on returning. References to child
        widgets (e.g., actions) of this window may be retained, however.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        '''

        # Initialize our superclass with all passed parameters.
        super().init(main_window)

        # Log this initialization.
        logs.log_debug('Sanitizing simulation configuration state...')

        # Initialize all widgets pertaining to the state of this simulation
        # configuration *BEFORE* connecting all relevant signals and slots
        # typically expecting these widgets to be initialized.
        self._init_widgets(main_window)
        self._init_connections(main_window)


    @type_check
    def _init_widgets(self, main_window: QBetseeMainWindow) -> None:
        '''
        Create all widgets owned directly by this object *and* initialize all
        other widgets (not necessarily owned by this object) whose internal
        state pertains to the high-level state of this simulation
        configuration.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow`
            widget.
        '''

        # Avoid circular import dependencies.
        from betsee.gui.simconf.guisimconfundo import (
            QBetseeUndoStackSimConf)

        # Classify all instance variables of this main window subsequently
        # required by this object. Since this main window owns this object,
        # since weak references are unsafe in a multi-threaded GUI context, and
        # since circular references are bad, this object intentionally does
        # *NOT* retain a reference to this main window.
        self._action_make_sim     = main_window.action_make_sim
        self._action_open_sim     = main_window.action_open_sim
        self._action_close_sim    = main_window.action_close_sim
        self._action_save_sim     = main_window.action_save_sim
        self._action_save_sim_as  = main_window.action_save_sim_as
        self._sim_conf_stack      = main_window.sim_conf_stack
        self._sim_conf_tree       = main_window.sim_conf_tree
        self._sim_conf_tree_frame = main_window.sim_conf_tree_frame
        self._sim_tab             = main_window.sim_tab

        # Undo stack for this simulation configuration.
        self.undo_stack = QBetseeUndoStackSimConf(
            main_window=main_window, sim_config=self)


    @type_check
    def _init_connections(self, main_window: QBetseeMainWindow) -> None:
        '''
        Connect all relevant signals and slots of *all* widgets (including the
        main window, top-level widgets of that window, and leaf widgets
        distributed throughout this application) whose internal state pertains
        to the high-level state of this simulation configuration.
        '''

        # Connect each such action to this object's corresponding slot.
        self._action_make_sim.triggered.connect(self._make_sim)
        self._action_open_sim.triggered.connect(self._open_sim)
        self._action_close_sim.triggered.connect(self._close_sim)
        self._action_save_sim.triggered.connect(self._save_sim)
        self._action_save_sim_as.triggered.connect(self._save_sim_as)

        # Connect this object's signals to all corresponding slots of *ALL*
        # objects across the codebase (including this object).
        #
        # Ideally, those objects would themselves (e.g., in their own init()
        # methods) connect these signals to these slots. In practice,
        # subsequent logic emits these signals and hence requires that these
        # connections be deterministically established *BEFORE* these signals
        # are emitted.
        self.set_filename_signal.connect(main_window.set_sim_conf_filename)

        #FIXME: Uncomment this *AFTER* defining the slot connected to here.
        # self.set_filename_signal.connect(main_window.sim_cmd.set_sim_conf_filename)
        self.set_filename_signal.connect(self.set_filename)
        self.set_dirty_signal.connect(main_window.set_sim_conf_dirty)
        self.set_dirty_signal.connect(self.set_dirty)

        # Set the state of all widgets dependent upon this simulation
        # configuration state *AFTER* connecting all relavant signals and
        # slots.  Initially, no simulation configuration has yet to be opened.
        #
        # Note that, as this slot only accepts strings, the empty string rather
        # than "None" is intentionally passed for safety.
        self.set_filename_signal.emit('')

    # ..................{ PROPERTIES ~ bool                 }..................
    @property
    def is_open(self) -> bool:
        '''
        ``True`` only if a simulation configuration file is currently open.
        '''

        return self.p.is_loaded

    # ..................{ PROPERTIES ~ str                  }..................
    @property
    def dirname(self) -> StrOrNoneTypes:
        '''
        Absolute path of the directory containing the currently open
        simulation configuration file if any *or* ``None`` otherwise.
        '''

        return self.p.conf_dirname


    @property
    def filename(self) -> StrOrNoneTypes:
        '''
        Absolute path of the currently open simulation configuration file if
        any *or* ``None`` otherwise.
        '''

        return self.p.conf_filename

    # ..................{ EXCEPTIONS                        }..................
    def die_unless_open(self) -> bool:
        '''
        Raise an exception unless a simulation configuration file is currently
        open.
        '''

        if not self.is_open:
            raise BetseeSimConfException(
                'No simulation configuration currently open.')

    # ..................{ SIGNALS                           }..................
    set_filename_signal = Signal(str)
    '''
    Signal passed either the absolute path of the currently open YAML-formatted
    simulation configuration file if any *or* the empty string otherwise.

    This signal is typically emitted on the user:

    * Opening a new simulation configuration.
    * Closing a currently open simulation configuration.
    '''


    set_dirty_signal = Signal(bool)
    '''
    Signal passed a single boolean on the currently open simulation
    configuration associated with the main window either receiving new unsaved
    changes (in which case this boolean is ``True``) *or* having just been
    saved (in which case this boolean is ``False``).

    This signal is typically emitted on each user edit of the contents of any
    widget owned by the top-level simulation configuration tree or stack
    widgets, implying a modification to this simulation configuration.
    '''

    # ..................{ SLOTS ~ state                     }..................
    @Slot(str)
    def set_filename(self, filename: str) -> None:
        '''
        Slot signalled on both the opening of a new simulation configuration
        and closing of an open simulation configuration.

        Parameters
        ----------
        filename : StrOrNoneTypes
            Absolute path of the currently open YAML-formatted simulation
            configuration file if any *or* the empty string otherwise (i.e., if
            no such file is open).
        '''

        # Notify all interested slots that no unsaved changes remain, regardless
        # of whether a simulation configuration has just been opened or closed.
        self.set_dirty_signal.emit(False)


    @Slot(bool)
    def set_dirty(self, is_dirty: bool) -> None:
        '''
        Slot signalled on each change of the **dirty state** (i.e., existence
        of unsaved in-memory changes) of the currently open simulation
        configuration if any.

        This slot internally updates the state (e.g., enabled or disabled,
        displayed or hidden) of all widgets owned or otherwise associated with
        this object.

        Parameters
        ----------
        is_dirty : bool
            ``True`` only if a simulation configuration is currently open *and*
            this configuration is **dirty** (i.e., has unsaved changes).
        '''

        # Classify this parameter.
        self._is_dirty = is_dirty

        # Enable or disable actions requiring an open simulation configuration.
        self._action_close_sim  .setEnabled(self.p.is_loaded)
        self._action_save_sim_as.setEnabled(self.p.is_loaded)

        # Enable or disable actions requiring an open simulation configuration
        # associated with an existing file and having unsaved changes.
        self._action_save_sim.setEnabled(self.p.is_loaded and is_dirty)

        # Show or hide widgets requiring an open simulation configuration.
        self._sim_conf_stack     .setEnabled(self.p.is_loaded)
        self._sim_conf_tree_frame.setEnabled(self.p.is_loaded)
        self._sim_tab            .setEnabled(self.p.is_loaded)

        # If no simulation configuration is open, clear the undo stack.
        if not self.p.is_loaded:
            self.undo_stack.clear()

    # ..................{ SLOTS ~ action                    }..................
    @Slot()
    def _make_sim(self) -> None:
        '''
        Slot invoked on the user requesting that the currently open simulation
        configuration if any be closed and a new simulation configuration with
        default settings be both created and opened.
        '''

        # Absolute path of a possibly non-existing YAML-formatted simulation
        # configuration file selected by the user.
        conf_filename = self._show_dialog_sim_conf_save()

        # If the user canceled this dialog, silently noop.
        if conf_filename is None:
            return
        # Else, the user did *NOT* cancel this dialog.

        # Close the currently open simulation configuration if any.
        self._close_sim()

        # Write the default simulation configuration to this file.
        confio.write_default(
            conf_filename=conf_filename,

            # Silently (over)write this file if this file already exists. Since
            # the guifile.select_file_save() function has already interactively
            # confirmed this overwrite in this case, doing so is safe to the
            # extent that the user has already accepted the consequences.
            is_conf_overwritable=True,

            # Preserve all external resources required by
            # this file with those contained in this default simulation
            # configuration. Since the guifile.select_file_save() function has
            # already interactively confirmed this overwrite when this file
            # already exists, doing so is safe to the extent that the user has
            # accepted the painful consequences.
        )

        # Deserialize this low-level file into a high-level configuration.
        self.load(conf_filename)

        # Update the status bar *AFTER* successfully completing this action.
        guiappstatus.show_status(QCoreApplication.translate(
            'QBetseeSimConf', 'Simulation created.'))


    @Slot()
    def _open_sim(self) -> None:
        '''
        Slot invoked on the user requesting that the currently open simulation
        configuration if any be closed and an existing external simulation
        configuration be opened.
        '''

        # Absolute path of an existing YAML-formatted simulation configuration
        # file selected by the user.
        conf_filename = self._show_dialog_sim_conf_open()

        # If the user canceled this dialog, silently noop.
        if conf_filename is None:
            return
        # Else, the user did *NOT* cancel this dialog.

        # Close the currently open simulation configuration if any.
        self._close_sim()

        # Deserialize this low-level file into a high-level configuration.
        self.load(conf_filename)

        # Update the status bar *AFTER* successfully completing this action.
        guiappstatus.show_status(QCoreApplication.translate(
            'QBetseeSimConf', 'Simulation opened.'))


    @Slot()
    def _close_sim(self) -> None:
        '''
        Slot invoked on the user attempting to close the currently open
        simulation configuration.

        If this configuration is dirty (i.e., has unsaved changes), the user
        will be interactively prompted to save this changes *before* this
        configuration is closed and these changes irrevocably lost.
        '''

        # If the user failed to interactively confirm saving all unsaved
        # changes if any for the currently open simulation configuration if
        # any, noop.
        if not self.save_if_dirty():
            return
        # Else, these change have all been saved.

        # Revert this configuration to the unloaded state.
        self.p.unload()

        # Notify all interested slots of this event.
        self.set_filename_signal.emit('')

        # Update the status bar *AFTER* successfully completing this action.
        guiappstatus.show_status(QCoreApplication.translate(
            'QBetseeSimConf', 'Simulation closed.'))


    @Slot()
    def _save_sim(self) -> None:
        '''
        Slot invoked on the user requesting all unsaved changes to the
        currently open simulation configuration be written to the external
        YAML-formatted file underlying this configuration, overwriting the
        contents of this file.
        '''

        # Reserialize this configuration back to the same file.
        self.p.save_inplace()

        # Notify all interested slots of this event.
        self.set_dirty_signal.emit(False)

        # Update the status bar *AFTER* successfully completing this action.
        guiappstatus.show_status(QCoreApplication.translate(
            'QBetseeSimConf', 'Simulation saved.'))


    @Slot()
    def _save_sim_as(self) -> None:
        '''
        Slot invoked on the user requesting the currently open simulation
        configuration be written to an arbitrary external YAML-formatted file.
        '''

        # Absolute path of a possibly non-existing YAML-formatted simulation
        # configuration file selected by the user.
        conf_filename = self._show_dialog_sim_conf_save()

        # If the user canceled this dialog, silently noop.
        if conf_filename is None:
            return
        # Else, the user did *NOT* cancel this dialog.

        # Reserialize this configuration into this new file.
        self.p.save(conf_filename)

        # Notify all interested slots of this event.
        self.set_filename_signal.emit(conf_filename)

        # Update the status bar *AFTER* successfully completing this action.
        guiappstatus.show_status(QCoreApplication.translate(
            'QBetseeSimConf', 'Simulation saved.'))

    # ..................{ LOADERS                           }..................
    @type_check
    def load(self, conf_filename: str) -> None:
        '''
        Deserialize the passed low-level YAML-formatted simulation configuration
        file into a high-level :class:`Parameters` object *and* signal all
        connected slots of this event.

        Design
        ----------
        Although low-level, this method is publicly accessible to permit the
        :class:`QBetseeMainWindow` class to handle the equally low-level
        ``--sim-conf-number`` command-line option.

        Note that, to avoid conflicts and confusion with ``open`` methods
        declared throughout the Qt API (e.g., :meth:`QDialog.open`,
        :meth:`QFile.open`), this method is intentionally *not* named ``open``.

        Parameters
        ----------
        conf_filename : str
            Absolute path of this file.
        '''

        # Deserialize this low-level file into a high-level configuration.
        self.p.load(conf_filename)

        # Signal all interested slots of this event.
        self.set_filename_signal.emit(conf_filename)

        # Focus the top-level tree widget *AFTER* signaling and hence populating
        # this tree widget to reflect the state of this configuration.
        self._sim_conf_tree.setFocus()

    # ..................{ SAVERS                            }..................
    def save_if_dirty(self) -> bool:
        '''
        Write all unsaved changes for the currently open simulation
        configuration to the external YAML-formatted file underlying this
        configuration if such a configuration is open, if this configuration is
        dirty (i.e., has unsaved changes), and if the user interactively
        confirms the overwriting of the existing contents of this file.

        Design
        ----------
        Although low-level, this method is publicly accessible to permit the
        :class:`QBetseeMainWindow` class to handle unsaved changes on window
        closure events.

        This method should typically be called immediately *before* the
        currently open simulation configuration (if any) is closed, preventing
        unsaved changes from being irrevocably lost.

        Returns
        ----------
        bool
            Either:

            * ``False`` only if a configuration is open, this configuration is
              dirty, and the user cancels the dialog prompting for
              confirmation.  In this case, the caller should ideally abort the
              current operation (e.g., closure of either the current window or
              simulation configuration).
            * ``True`` in *all* other cases.
        '''

        # If this configuration is *NOT* dirty (i.e., has unsaved changes),
        # report success as no changes remain to be saved.
        if not self._is_dirty:
            return True
        # Else, this configuration is dirty.

        # Interactively prompt the user to save these changes and store the
        # bit value of the "QMessageBox.StandardButton" enumeration member
        # signifying the button clicked by the user.
        button_clicked = guimessage.show_warning(
            title=QCoreApplication.translate(
                'QBetseeSimConf', 'Unsaved Simulation Configuration'),
            synopsis=QCoreApplication.translate(
                'QBetseeSimConf',
                'The currently open simulation configuration has '
                'unsaved changes.'
            ),
            exegesis=QCoreApplication.translate(
                'QBetseeSimConf', 'Would you like to save these changes?'),
            buttons=(
                QMessageBox.Save |
                QMessageBox.Discard |
                QMessageBox.Cancel
            ),
            button_default=QMessageBox.Save,
        )

        # If the "Cancel" button was clicked, report failure.
        if button_clicked == QMessageBox.Cancel:
            return False

        # If the "Save" button was clicked, save these changes. Namely,
        # reserialize this configuration back to the same file.
        if button_clicked == QMessageBox.Save:
            self.p.save_inplace()
        # Else, the "Discard" button was clicked. Discard these changes by
        # doing absolutely nothing.

        # In either case, report success.
        return True

    # ..................{ SHOWERS                           }..................
    def _show_dialog_sim_conf_open(self) -> str:
        '''
        Display a dialog requiring the user to select an existing
        YAML-formatted file to be subsequently opened for reading (rather than
        overwriting) as the new simulation configuration, returning the
        absolute path of this file if this dialog was not canceled *or*
        ``None`` otherwise (i.e., if this dialog was canceled).
        '''

        return guifile.select_file_yaml_read(
            dialog_title=QCoreApplication.translate(
                'QBetseeSimConf', 'Open Simulation Configuration'))


    def _show_dialog_sim_conf_save(self) -> str:
        '''
        Display a dialog requiring the user to select a YAML-formatted file
        (either existing or non-existing) to be subsequently opened for
        in-place saving and hence overwriting as the new simulation
        configuration, returning the absolute path of this file if this dialog
        was not canceled *or* ``None`` otherwise (i.e., if this dialog was
        canceled).
        '''

        return guifile.select_file_yaml_save(
            dialog_title=QCoreApplication.translate(
                'QBetseeSimConf', 'New Simulation Configuration'))



#!/usr/bin/env python3
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level :mod:`QSettings`-based application-wide slottable settings classes.
'''

# ....................{ IMPORTS                            }....................
from PySide2.QtCore import QObject, Slot
from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.gui.window.guimainwindow import QBetseeMainWindow
from betsee.util.io import guisettings

# ....................{ CLASSES                            }....................
class QBetseeSettings(QObject):
    '''
    :class:`QSettings`-based object exposing all application-wide settings with
    cross-platform, thread- and process-safe slots permitting external callers
    to request restoration and storage of these settings to and from their
    on-disk backing store (e.g., an application- and user-specific dotfile).

    Attributes
    ----------
    _main_window : QBetseeMainWindow
        Main window for this application.

    See Also
    ----------
    :class:`betsee.gui.guisignal.QBetseeSignaler`
        Sibling class, whose signals are connected to this object's slots.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, main_window: QBetseeMainWindow, *args, **kwargs) -> None:
        '''
        Initialize this slotter, connecting each slot to the corresponding
        signal of the :class:`QBetseeSettingsSignaler` instance owned by the
        passed main window widget.

        Parameters
        ----------
        main_window : QBetseeMainWindow
            Initialized application-specific parent :class:`QMainWindow` widget
            against which to initialize this object.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Classify all remaining parameters.
        self._main_window = main_window

        # Signaler owned by this window.
        signaler = self._main_window.signaler

        # Connect this signaler to this slotter.
        signaler.restore_settings_signal.connect(self.restore_settings)
        signaler.store_settings_signal.connect(self.store_settings)

    # ..................{ SLOTS                              }..................
    #FIXME: This should be called:
    #
    #* By the QGuiApplication::saveStateRequest() slot, which Qt triggers on
    #  the current desktop session manager beginning a restoration from a prior
    #  shutdown or suspend.
    @Slot()
    def restore_settings(self) -> None:
        '''
        Read and restore application-wide settings previously written to a
        predefined application- and user-specific on-disk file by the most
        recent execution of this application if any *or* reduce to a noop.
        '''

        #FIXME: Excise this.
        # return

        # Log this restoration.
        logs.log_info('Restoring application settings...')

        # Previously written application settings.
        settings = guisettings.make()

        # Read settings specific to this main window.
        settings.beginGroup('MainWindow')

        # Restore all previously written properties of this window.
        #
        # Note that there exist numerous means of doing so. While the canonical
        # means of doing so appears to be the QMainWindow.restoreGeometry() and
        # QMainWindow.restoreState() methods, QSettings documentation explicitly
        # states that:
        #
        #     "See Window Geometry for a discussion on why it is better to call
        #      QWidget::resize() and QWidget::move() rather than
        #      QWidget::setGeometry() to restore a window's geometry."
        #
        # Sadly, the "Window Geometry" article fails to actually discuss why the
        # QWidget.resize() and QWidget.move() methods are preferable to
        # QWidget.setGeometry() with respect to the main window. We do note,
        # however, that QWidget.setGeometry() documentation cautions:
        #
        #     "Warning: Calling setGeometry() inside resizeEvent() or
        #      moveEvent() can lead to infinite recursion."
        #
        # In the absence of compelling evidence, the current approach prevails.

        #FIXME: This logic currently appears to produce a bizarrely unusable
        #window on some platforms (e.g., Ubuntu Linux), presumably due to the
        #QBetseeMainWindow.resize_full() method and hence Qt.WindowFullScreen
        #mode being fundamentally broken in some respect -- possibly due to
        #Qt 5.6.x-specific issues resolved in subsequent Qt versions. In any
        #event, this logic *MUST* be disabled for the moment.

        # If a preferred window full screen state was previously stored...
        # if settings.contains('isFullScreen'):
        #     # Previously stored window full-screen state.
        #     main_window_is_full_screen = bool(settings.value('isFullScreen'))
        #
        #     # Log this restoration.
        #     logs.log_debug(
        #         'Restoring window full-screen state "%r"...',
        #         main_window_is_full_screen)
        #
        #     # If restoring this window to a full-screen state, do so.
        #     if main_window_is_full_screen:
        #         self._main_window.resize_full()

        # If a preferred window position was previously stored...
        if settings.contains('pos'):
            # Previously stored window position.
            main_window_position = settings.value('pos')

            # Log this restoration.
            logs.log_debug(
                'Restoring window position %r...', main_window_position)

            # Restore this position.
            self._main_window.move(main_window_position)
        # Else, position this window at the origin -- ensuring that subsequently
        # maximizing this window consumes all available screen space.
        else:
            self._main_window.move(0, 0)

        # If a preferred window size was previously stored...
        if settings.contains('size'):
            # Previously stored window size.
            main_window_size = settings.value('size')

            # Log this restoration.
            logs.log_debug('Restoring window size %r...', main_window_size)

            # Restore this size.
            self._main_window.resize(main_window_size)
        # Else, expand this window to consume all available screen space in a
        # non-"full screen" manner. Failing to do so often results in Qt
        # defaulting this window to an inappropriate size for the current
        # screen. On my 1920x1080 display, for example, the default window size
        # exceeds the vertical resolution (and hence is clipped at the bottom)
        # but consumes only two-thirds of the horizontal resolution.
        else:
            self._main_window.resize_max()

        # Cease reading settings specific to this window.
        settings.endGroup()


    #FIXME: This should be called:
    #
    #* Ideally incrementally during the application life cycle to prevent
    #  settings from being lost if the application fails to close gracefully.
    #  "QTimer" is probably our friends, here.
    #* By the QGuiApplication::commitDataRequest() slot, which Qt triggers on
    #  the current desktop session manager beginning a shutdown or suspend.
    #  However, note that:
    #  * This slot will also need to save unsaved data if any and that no
    #    interactive message box should be displayed to the user *UNLESS* this
    #    session manager explicitly permits such interaction. (Sigh.)
    #  * The QGuiApplication.setFallbackSessionManagementEnabled(False) method
    #    will need to be called to prevent fallback session management from
    #    interfering with this slot's behaviour.
    #  * Any concurrent operations (e.g., simulation running) will need to be
    #    temporarily halted until this application is restored. Failure to do so
    #    typically results in the OS killing this application. (Makes sense.) To
    #    respond to this application state change in a robust manner, it will
    #    probably be necessary to connect to the
    #    QGuiApplication::applicationStateChanged() slot.
    @Slot()
    def store_settings(self) -> None:
        '''
        Write application-wide settings to a predefined application- and
        user-specific on-disk file, which the next execution of this application
        will read and restore on startup.
        '''

        # Log this storage.
        logs.log_info('Storing application settings...')

        # Currently written application settings if any.
        settings = guisettings.make()

        # Write settings specific to this main window.
        settings.beginGroup('MainWindow')

        # Current window full-screen state, position, and size.
        main_window_is_full_screen = self._main_window.isFullScreen()
        main_window_position = self._main_window.pos()
        main_window_size = self._main_window.size()

        # Log these window properties.
        logs.log_debug('Storing window full-screen state "%r"...',
                       main_window_is_full_screen)
        logs.log_debug('Storing window position %r...', main_window_position)
        logs.log_debug('Storing window size %r...', main_window_size)

        # Store these window properties.
        settings.setValue('isFullScreen', main_window_is_full_screen)
        settings.setValue('pos', main_window_position)
        settings.setValue('size', main_window_size)

        # Cease writing settings specific to this window.
        settings.endGroup()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
General-purpose :mod:`QAbstractSpinBox` widget subclasses.
'''

# ....................{ IMPORTS                            }....................
# from PySide2.QtCore import Signal, Slot
from PySide2.QtGui import QValidator
from PySide2.QtWidgets import QDoubleSpinBox
from betse.util.type.numeric import floats
from betse.util.type.text import regexes
from betse.util.type.types import type_check

# ....................{ SUBCLASSES                         }....................
class QBetseeDoubleSpinBox(QDoubleSpinBox):
    '''
    :mod:`QDoubleSpinBox`-based widget optimized for intelligent display of
    floating point numbers.

    This application-specific widget augments the stock :class:`QDoubleSpinBox`
    widget with additional support for scientific notation, permitting *only*
    characters permissible in both:

    * Decimal notation (e.g., digits, signs, and the radix point).
    * Scientific notation (e.g., digits, signs, the radix point, and the letter
      "e" in both capitalized and uncapitalized variants).

    Attributes
    ----------
    _validator : QBetseeDoubleValidator
        Application-specific validator validating floating point numbers in both
        decimal and scientific notation.

    See Also
    ----------
    https://jdreaver.com/posts/2014-07-28-scientific-notation-spin-box-pyside.html
        Blog post partially inspiring this implementation.
    '''

    # ..................{ CONSTANTS                          }..................
    SINGLE_STEP_DEFAULT = 1.0
    '''
    Default Qt value for the :meth:`singleStep` property.

    Testing the current value of the :meth:`singleStep` property against this
    default permits subclasses to determine whether a widget-specific value for
    this property has been specified or not.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Connect the text appending signal to the corresponding slot.
        self._validator = QBetseeDoubleValidator()

        # Prevent Qt from constraining displayed floating point numbers to an
        # arbitrary maximum of 99.99, an insane default with no correspondence
        # to actual reality.
        #
        # By contrast, we silently permit Qt to retains its default behaviour of
        # constraining displayed floating point numbers to a minimum of 0, a
        # sane default given the inability of most measurable physical
        # quantities to "go negative."
        #
        # Note that overriding the default maximum here still permits callers to
        # override this with a widget- or subclass-specific maximum. Hence, this
        # override has *NO* unintended or harmful side effects.
        self.setMaximum(floats.FLOAT_MAX)

    # ..................{ VALIDATORS                         }..................
    # See the "QBetseeDoubleValidator" class for commentary.

    def validate(self, text: str, char_index: int) -> QValidator.State:
        return self._validator.validate(text, char_index)

    def fixup(self, text: str) -> str:
        return self._validator.fixup(text)

    # ..................{ VALIDATORS                         }..................
    def valueFromText(self, text: str) -> float:
        '''
        Floating point number converted from the passed human-readable string.
        '''

        return float(text)


    def textFromValue(self, number: float) -> str:
        '''
        Human-readable string converted from the passed floating point number.
        '''

        # For readability, a high-level utility function wrapping the low-level
        # "{:g}" format specifier is preferred.
        return floats.to_str(number)

# ....................{ SUBCLASSES ~ validator             }....................
class QBetseeDoubleValidator(QValidator):
    '''
    Validator enabling the :mod:`QBetseeDoubleSpinBox` widget to input and
    display floating point numbers in both decimal and scientific notation.

    This application-specific widget augments the stock
    :class:`QDoubleValidator` validator with additional support for scientific
    notation.

    Attributes
    ----------
    _float_regex : RegexCompiledType
        Compiled regular expression matching a floating point number.

    See Also
    ----------
    https://jdreaver.com/posts/2014-07-28-scientific-notation-spin-box-pyside.html
        Blog post partially inspiring this implementation.
    '''

    # ..................{ CONSTANTS                          }..................
    _CHARS_NONDIGIT = {'e', 'E', '.', '-', '+'}
    '''
    Set of all permissible non-digit characters in floating point numbers in
    either decimal or scientific notation.
    '''

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Initialize our superclass with all passed parameters.
        super().__init__(*args, **kwargs)

        # Compiled regular expression matching a floating point number in any
        # format, cached to ignorably improve efficiency in the frequently
        # called fixup() method.
        self._float_regex = floats.get_float_regex()

    # ..................{ VALIDATORS                         }..................
    @type_check
    def validate(self, text: str, char_index: int) -> QValidator.State:
        '''
        Validate the passed input text at the passed character index of this
        text to be a valid floating point number or not.

        Specifically, this method returns:

        * :attr:`State.Acceptable` if this text is guaranteed to be an valid
          floating point number.
        * :attr:`State.Intermediate` if this text is currently an invalid
          floating point number but could conceivably be transformed into a
          valid floating point number by subsequent input of one or more
          printable characters.
        * :attr:`State.Invalid` if this text is guaranteed to be an invalid
          floating point number regardless of subsequently input printable
          characters.

        Parameters
        ----------
        text : str
            Input text to be validated.
        char_index : int
            0-based character index of the input cursor in this text.

        Returns
        ----------
        State
            Whether this text is a valid floating point number or not.
        '''

        # If this text is a valid floating point number, inform the caller.
        if floats.is_float_str(text):
            return QValidator.State.Acceptable
        # Else, this text is *NOT* a valid floating point number.
        #
        # If this text is either the empty string *OR* the character
        # preceding the current character in this text is a permissible
        # non-digit in either decimal or scientific notation, this text could
        # conceivably be transformed into a valid floating point number by
        # subsequent entry of one or more printable characters.
        elif text == '' or text[char_index-1] in self._CHARS_NONDIGIT:
            return QValidator.State.Intermediate
        # Else, this text is definitively an invalid floating point number.
        else:
            return QValidator.State.Invalid


    @type_check
    def fixup(self, text: str) -> str:
        '''
        Input guaranteed to be in either the :attr:`State.Intermediate` or
        :attr:`State.Valid` states, sanitized from the passed input guaranteed
        to be in the :attr:`State.Invalid` state.

        Specifically, this function returns:

        * The first substring in this invalid input matching a floating point
          number in either decimal or scientific notation if this input contains
          such a substring. In this case, the returned string is valid.
        * The empty string otherwise. In this case, the returned string is
          merely intermediate.

        In either case, the returned string is guaranteed *not* to be invalid.

        Parameters
        ----------
        text : str
            Invalid input text to be munged.

        Returns
        ----------
        str
            Intermediate or valid input munged from this invalid input.
        '''

        return regexes.get_match_full_first_if_any(
            text=text, regex=self._float_regex) or ''



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **signals-based simulation phase callbacks** (i.e., collection of all
simulation phase callbacks whose methods emit queued Qt signals) classes.
'''

# ....................{ IMPORTS                           }....................
# from PySide2.QtCore import QCoreApplication  # Slot, Signal
from betse.science.phase.phasecallbacks import SimCallbacksBC
# from betse.util.io.log import logs
from betse.util.type.types import type_check
from betsee.util.thread.pool.guipoolworksig import (
    QBetseeThreadPoolWorkerSignals)

# ....................{ SUBCLASSES ~ callbacks            }....................
class SimCallbacksSignaller(SimCallbacksBC):
    '''
    **Signals-based simulation phase callbacks** (i.e., caller-defined object
    whose methods emit signals on the :class:`QBetseeThreadPoolWorkerSignals`
    object with which this :class:`SimCallbacksSignaller` object is
    initializedare on being periodically called while simulating one or more
    simulation phases).

    This object effectively glues low-level simulation subcommands to
    high-level simulator widgets, converting callbacks called by the former
    into signals emitted on slots defined by the latter.

    Attributes
    ----------
    _signals : QBetseeThreadPoolWorkerSignals
        Collection of all signals emittable by simulator workers.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, signals: QBetseeThreadPoolWorkerSignals) -> None:
        '''
        Initialize this callbacks collection.

        Parameters
        ----------
        signals : QBetseeThreadPoolWorkerSignals
            Collection of all signals emittable by simulator workers.
        '''

        # Initialize our superclass with all passed parameters.
        super().__init__()

        # Classify all passed parameters.
        self._signals = signals

    # ..................{ CALLBACKS ~ progress              }..................
    @type_check
    def progress_ranged(
        self, progress_max: int, progress_min: int = 0) -> None:

        # Perform all superclass callback handling first.
        super().progress_ranged(
            progress_min=progress_min, progress_max=progress_max)

        # Forward these callback parameters to the corresponding worker signal.
        self._signals.emit_progress_range(
            progress_min=progress_min, progress_max=progress_max)


    @type_check
    def progress_stated(self, status: str) -> None:

        # Perform all superclass callback handling first.
        super().progress_stated(status=status)

        # Forward these callback parameters to the corresponding worker signal.
        self._signals.emit_progress_state(status=status)


    @type_check
    def progressed(self, progress: int) -> None:

        # Perform all superclass callback handling first.
        super().progressed(progress=progress)

        # Forward these callback parameters to the corresponding worker signal.
        self._signals.emit_progress(progress=progress)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Metadata constants synopsizing high-level application dependencies.
'''

# ....................{ IMPORTS                           }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid race conditions during setuptools-based installation, this
# module may import *ONLY* from modules guaranteed to exist at the start of
# installation. This includes all standard Python and application modules but
# *NOT* third-party dependencies, which if currently uninstalled will only be
# installed at some later time in the installation.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betsee.guimetadata import VERSION

# ....................{ LIBS ~ runtime : mandatory        }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: Changes to this subsection *MUST* be synchronized with:
# * Front-facing documentation (e.g., "doc/md/INSTALL.md").
# * The "betse.util.type.modules.DISTUTILS_PROJECT_NAME_TO_MODULE_NAME"
#   dictionary, converting between the setuptools-specific names listed below
#   and the Python-specific module names imported by this application.
# * Gitlab-CI configuration (e.g., the top-level "requirements-conda.txt" file).
# * Third-party platform-specific packages (e.g., Gentoo Linux ebuilds).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# See the "Design" section below for commentary.
BETSE_VERSION_REQUIRED_MIN = VERSION[:VERSION.rindex('.')]
'''
Minimum version of BETSE, the low-level CLI underlying this high-level GUI,
required by this application as a human-readable ``.``-delimited string.

Whereas all other minimum versions of third-party dependencies required by this
application are specified as key-value pairs of various dictionary globals of
this submodule, this minimum version is specified as an independent global --
simplifying inspection and validation of this version elsewhere (e.g., in the
:func:`betsee.__main__.die_unless_betse` function).

Design
----------
By design, the current version of this application requires at least the same
version of BETSE ignoring the trailing component of the former -- an
application-specific patch number allowing each version of BETSE to satisfy
multiple versions of this application. For example, BETSE 0.8.1 satisfies both
versions 0.8.1.0 *and* 0.8.1.1 of this application.
'''


RUNTIME_MANDATORY = {
    # Versioned dependencies directly required by this application.
    'BETSE': '>= ' + BETSE_VERSION_REQUIRED_MIN,
    'PySide2': '>= 2.0.0~alpha0',

    # Unversioned dependencies directly required by this application. Since
    # the modules providing these dependencies define no PEP-8-compliant
    # "__version__" or "__version_info__" attributes. merely validating these
    # modules to be importable is the most we can do.
    'PySide2.QtGui': '',
    'PySide2.QtSvg': '',
    'PySide2.QtWidgets': '',
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
mandatory runtime dependency for this application to the suffix of a
:mod:`setuptools`-specific requirements string constraining this dependency.

See Also
----------
:func:`get_runtime_mandatory_sans_submodules`
    Function returning a copy of this dictionary excluding all :mod:`PySide2`
    submodules (e.g., :mod:`PySide2.QtGui`).
:data:`betse.metadata.RUNTIME_MANDATORY`
    Further details on dictionary structure.
:download:`/doc/rst/INSTALL.rst`
    Human-readable list of these dependencies.
'''

# ....................{ LIBS ~ runtime : optional         }....................
#FIXME: Should these dependencies also be added to our "setup.py" metadata,
#perhaps as so-called "extras"? Contemplate. Consider. Devise.
RUNTIME_OPTIONAL = {
    # To simplify subsequent lookup at runtime, project names for optional
    # dependencies should be *STRICTLY LOWERCASE*. Since setuptools parses
    # project names case-insensitively, case is only of internal relevance.

    #FIXME: Add a minimum required version *AFTER* upstream resolves the
    #following open issue:
    #    https://bugreports.qt.io/browse/PYSIDE-517
    #FIXME: The official "PySide2" wheel now ships "pyside2uic" out-of-the-box,
    #suggesting this should now resemble:
    #    'pyside2uic': RUNTIME_MANDATORY['PySide2'],
    #Test the above specification when time admits (i.e., sadly never).
    'pyside2uic': '',
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
optional runtime dependency for this application to the suffix of a
:mod:`setuptools`-specific requirements string constraining this dependency.

See Also
----------
:data:`betse.metadata.RUNTIME_MANDATORY`
    Further details on dictionary structure.
:download:`/doc/rst/INSTALL.rst`
    Human-readable list of these dependencies.
'''

# ....................{ LIBS ~ testing : mandatory        }....................
TESTING_MANDATORY = {
    # For simplicity, py.test should remain the only hard dependency for
    # testing on local machines. While our setuptools-driven testing regime
    # optionally leverages third-party py.test plugins (e.g., "pytest-xdist"),
    # these plugins are *NOT* required for simple testing.
    'pytest': '>= 2.5.0',
}
'''
Dictionary mapping from the :mod:`setuptools`-specific project name of each
mandatory testing dependency for this application to the suffix of a
:mod:`setuptools`-specific requirements string constraining this dependency.

See Also
----------
:data:`betse.metadata.RUNTIME_MANDATORY`
    Further details on dictionary structure.
:download:`/doc/rst/INSTALL.rst`
    Human-readable list of these dependencies.
'''

# ....................{ GETTERS                           }....................
def get_runtime_mandatory_tuple() -> tuple:
    '''
    Tuple listing the :mod:`setuptools`-specific requirement string containing
    the mandatory name and optional version and extras constraints of each
    mandatory runtime dependency for this application, dynamically converted
    from the :data:`metadata.RUNTIME_MANDATORY` dictionary.

    Caveats
    ----------
    This dictionary notably excludes all submodules whose fully-qualified names
    are prefixed by ``PySide2.`` (e.g., :mod:`PySide2.QtGui`). These submodules
    signify optional :mod:`PySide2` components required by this application but
    unavailable on PyPI. Including these submodules here would erroneously halt
    setuptools-based installation for up to several minutes with output
    resembling:

        Searching for PySide2.QtSvg
        Reading https://pypi.python.org/simple/PySide2.QtSvg/
        Couldn't find index page for 'PySide2.QtSvg' (maybe misspelled?)
        Scanning index of all packages (this may take a while)
        Reading https://pypi.python.org/simple/
    '''

    # Avoid circular import dependencies.
    from betsee.lib.setuptools import guisetuptool

    # Dictionary of all mandatory runtime dependencies excluding submodules.
    runtime_mandatory_sans_submodules = {
        # Map this dependency's name to constraints.
        dependency_name: dependency_constraints

        # For the name and constraints of each mandatory runtime dependency...
        for dependency_name, dependency_constraints in
            RUNTIME_MANDATORY.items()

        # If this is neither a PySide2-specific submodule nor the "pyside2uic"
        # subpackage, which official PySide2 wheels now bundle out-of-the-box
        # and hence are *NOT* externally available from PyPI...
        if not dependency_name.startswith(('PySide2.', 'pyside2uic'))
    }

    # Return this dictionary converted into a tuple.
    return guisetuptool.convert_requirements_dict_to_tuple(
        runtime_mandatory_sans_submodules)


def get_runtime_optional_tuple() -> tuple:
    '''
    Tuple listing the :mod:`setuptools`-specific requirement string containing
    the mandatory name and optional version and extras constraints of each
    optional runtime dependency for this application, dynamically converted
    from the :data:`metadata.RUNTIME_OPTIONAL` dictionary.
    '''

    # Avoid circular import dependencies.
    from betsee.lib.setuptools import guisetuptool

    # Return this dictionary converted into a tuple.
    return guisetuptool.convert_requirements_dict_to_tuple(RUNTIME_OPTIONAL)


def get_testing_mandatory_tuple() -> tuple:
    '''
    Tuple listing the :mod:`setuptools`-specific requirement string containing
    the mandatory name and optional version and extras constraints of each
    mandatory testing dependency for this application, dynamically converted
    from the :data:`metadata.TESTING_MANDATORY` dictionary.
    '''

    # Avoid circular import dependencies.
    from betsee.lib.setuptools import guisetuptool

    # Return this dictionary converted into a tuple.
    return guisetuptool.convert_requirements_dict_to_tuple(TESTING_MANDATORY)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2017-2018 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **simulator worker enumeration** (e.g., :class:`enum.Enum` subclass
describing different types of simulator work) functionality.
'''

# ....................{ IMPORTS                           }....................
from betsee.gui.simtab.run.guisimrunenum import SimmerState
from enum import Enum

# ....................{ ENUMERATIONS                      }....................
class SimmerPhaseSubkind(Enum):
    '''
    Enumeration of each kind of **simulator subcommand worker** (i.e., type of
    work performed within a given simulation phase by a given simulator
    worker).

    This child enumeration is a proper subset of the parent
    :class:`SimmerState` enumeration differentiating simulator workers from one
    another in a convenient manner permitting the state of each simulator phase
    acted upon by each such worker to be trivially set.

    Attributes
    ----------
    MODELLING : enum
        Modelling state, implying this worker to model its simulator phase.
    EXPORTING : enum
        Exporting state, implying this worker to export its simulator phase.

    Examples
    ----------
    To trivially map between members of these two correlated enumerations, each
    member of this child enumeration is guaranteed to have the same ``value``
    attribute as each corresponding member of the parent :class:`SimmerState`
    enumeration: e.g.,

        >>> from betsee.gui.simtab.run.guisimrunstate import SimmerState
        >>> from betsee.gui.simtab.run.work.guisimrunworkenum import (
        ...     SimmerPhaseSubkind)
        >>> SimmerPhaseSubkind.MODELLING.value == (
        ...     SimmerState.MODELLING.value)
        True
        >>> SimmerPhaseSubkind.EXPORTING.value == (
        ...     SimmerState.EXPORTING.value)
        True
    '''

    # For simplicity elsewhere, reuse existing enumeration values.
    MODELLING = SimmerState.MODELLING.value
    EXPORTING = SimmerState.EXPORTING.value
